"""Utility functions for synthesizing MIDI."""

#  Copyright 2021 The DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  Lint as: python3
#
#  Lint as: python3

import ddsp.core
import pretty_midi
import numpy as np
import tensorflow as tf
import os
from midi_ddsp.data_handling.instrument_name_utils import INST_ABB_TO_ID_DICT
from midi_ddsp.utils.audio_io import save_wav
from midi_ddsp.utils.inference_utils import ensure_same_length, \
  expression_generator_output_to_conditioning_df, \
  conditioning_df_to_dict, conditioning_df_to_audio, \
  conditioning_df_to_midi_features, CONDITIONING_KEYS


# MIDI Synthesis

def time_to_frame(time_s, fs=250):
  """Convert time in second to frame"""
  return int(round(time_s * fs))


def note_list_to_sequence(note_list, fs=250, pitch_offset=0, speed_rate=1,
                          remove_start_silence=False):
  """Convert list of note in pretty_midi to midi sequence for
  expression generator."""
  note_pitch = []
  note_length = []

  prev_off_frame = 1000000
  for note in note_list:
    on_frame = time_to_frame(note.start / speed_rate, fs)
    off_frame = time_to_frame(note.end / speed_rate, fs)
    if on_frame - prev_off_frame > 0:
      note_pitch.append(0)
      note_length.append(on_frame - prev_off_frame)
    note_pitch.append(note.pitch + pitch_offset)
    note_length.append(off_frame - on_frame)
    prev_off_frame = off_frame

  # add the silence to the start if the remove_start_silence is False
  if not remove_start_silence:
    note_pitch.insert(0, 0)
    note_length.insert(0, time_to_frame(note_list[0].start / speed_rate, fs))

  # add a rest of 1 second to the end for better synthesis quality
  note_pitch.append(0)
  note_length.append(fs)

  # adding batch dimension
  note_pitch = tf.convert_to_tensor(np.array(note_pitch))[tf.newaxis, ...]
  note_length = tf.convert_to_tensor(np.array(note_length))[tf.newaxis, ...]

  note_length = tf.cast(note_length, tf.float32)[..., tf.newaxis] * 0.004

  note_sequence = {'note_pitch': note_pitch, 'note_length': note_length}
  return note_sequence


def mono_midi_to_note_sequence(midi_file, instrument_id, pitch_offset=0,
                               speed_rate=1):
  """Convert a mono MIDI file to note sequence for expression generator."""
  midi_data = pretty_midi.PrettyMIDI(midi_file)
  instrument = midi_data.instruments[0]
  if instrument.is_drum:
    raise ValueError('Cannot synthesize drum')
  note_sequence = note_list_to_sequence(instrument.notes, fs=250,
                                        pitch_offset=pitch_offset,
                                        speed_rate=speed_rate)
  note_sequence['instrument_id'] = instrument_id
  return note_sequence


# Macro dicts for ensembles.
STRING_SET = {
  'Soprano': 'vn',
  'Alto': 'vn',
  'Tenor': 'va',
  'Bass': 'vc',
}

WOODWIND_SET = {
  'Soprano': 'fl',
  'Alto': 'ob',
  'Tenor': 'cl',
  'Bass': 'bn',
}

BRASSWIND_SET = {
  'Soprano': 'tpt',
  'Alto': 'hn',
  'Tenor': 'tbn',
  'Bass': 'tba',
}

MIX_SET = {
  'Soprano': 'cl',
  'Alto': 'tpt',
  'Tenor': 'sax',
  'Bass': 'tba',
}

QUARTET_SET_DICT = {
  'string_set': STRING_SET,
  'woodwind_set': WOODWIND_SET,
  'brasswind_set': BRASSWIND_SET,
  'mix_set': MIX_SET
}


def bach_midi_to_features(midi_file, quartet_set='woodwind_set',
                          quartet_set_dict=None, pitch_offset=0,
                          speed_rate=1):
  """Convert Bach chorales MIDI file to features for expression generator."""
  if quartet_set_dict is None:
    quartet_set_dict = QUARTET_SET_DICT

  midi_data = pretty_midi.PrettyMIDI(midi_file)
  note_sequence_all = []
  instrument_id_all = []
  part_name_all = []
  for instrument in midi_data.instruments:
    if instrument.is_drum:
      raise ValueError('Cannot synthesize drum')
    part_name = instrument.name.rstrip('\x00')
    note_sequence = note_list_to_sequence(instrument.notes, fs=250,
                                          pitch_offset=pitch_offset,
                                          speed_rate=speed_rate)
    instrument_abb = quartet_set_dict[quartet_set][part_name]
    instrument_id = tf.constant([INST_ABB_TO_ID_DICT[instrument_abb]])
    note_sequence['instrument_id'] = instrument_id
    instrument_id_all.append(instrument_id)
    note_sequence_all.append(note_sequence)
    part_name_all.append(part_name)
  return note_sequence_all, instrument_id_all, part_name_all


def synthesize_bach(synthesis_generator, expression_generator, midi_file,
                    quartet_set='string_set', pitch_offset=0,
                    speed_rate=1, output_dir=r'./', gain_adjust_db_dict=None):
  """Synthesize Bach chorales MIDI."""

  # Name for four parts.
  part_name_all = ['Soprano',
                   'Alto',
                   'Tenor',
                   'Bass', ]

  midi_data = pretty_midi.PrettyMIDI(midi_file)
  if len(midi_data.instruments) != 4:
    raise ValueError('The MIDI file do not have four parts')
  for i, instrument in enumerate(midi_data.instruments):
    instrument.name = part_name_all[i]
  midi_data.write(midi_file)

  # Create output directory.
  piece_name = os.path.basename(midi_file).replace('.mid', '')
  save_dir = os.path.join(output_dir, f'{piece_name}_{quartet_set}')
  os.makedirs(save_dir, exist_ok=True)

  # Synthesize the audio.
  note_sequence_all, instrument_id_all, part_name_all = bach_midi_to_features(
    midi_file, quartet_set=quartet_set,
    pitch_offset=pitch_offset,
    speed_rate=speed_rate)

  midi_audio_all = []
  conditioning_df_all = []
  for note_sequence, instrument_id in zip(note_sequence_all, instrument_id_all):
    expression_generator_outputs = expression_generator(note_sequence, out=None,
                                                        training=False)
    conditioning_df = expression_generator_output_to_conditioning_df(
      expression_generator_outputs['output'], note_sequence)
    conditioning_df_all.append(conditioning_df)

  midi_audio, midi_control_params, midi_synth_params = \
    batch_conditioning_df_to_audio(
      synthesis_generator,
      conditioning_df_all,
      instrument_id_all)

  # Save the stem audios.
  for i in range(midi_audio.shape[0]):
    midi_audio_all.append(midi_audio[i].numpy())
    save_wav(midi_audio[i].numpy(),
             os.path.join(save_dir, f'{part_name_all[i]}.wav'), 16000)

  # Mix the stems.
  if not gain_adjust_db_dict:
    gain_adjust_db_dict = {part_name: 0 for part_name in part_name_all}

  for i in range(len(midi_audio_all)):
    gain_adjust_amp_dict = {k: 10 ** (v / 20) for k, v in
                            gain_adjust_db_dict.items()}
    gain = gain_adjust_amp_dict[part_name_all[i]]
    midi_audio_all[i] = midi_audio_all[i] * gain

  # Save the mix.
  midi_audio_mix = tf.reduce_sum(tf.stack(midi_audio_all, axis=-1),
                                 axis=-1).numpy()
  save_wav(midi_audio_mix, os.path.join(save_dir, 'mix.wav'), 16000)

  return midi_audio_mix, midi_audio_all, midi_control_params, \
         midi_synth_params, conditioning_df_all


def batch_conditioning_df_to_audio(synthesis_generator, conditioning_df_all,
                                   instrument_id_all,
                                   display_progressbar=True):
  """Generate audio from a batch of conditioning_df."""
  conditioning_dict_all = [conditioning_df_to_dict(conditioning_df) for
                           conditioning_df in conditioning_df_all]
  conditioning_dict_all_concat = {}
  for key in conditioning_dict_all[0].keys():
    conditioning_dict_all_concat[key] = tf.concat(
      ensure_same_length([c[key] for c in conditioning_dict_all]), 0)
  conditioning_dict = conditioning_dict_all_concat

  midi_features_all = [conditioning_df_to_midi_features(conditioning_df) for
                       conditioning_df in conditioning_df_all]
  midi_features_all_concat = []
  for i in range(len(midi_features_all[0])):
    midi_features_all_concat.append(
      tf.concat(ensure_same_length([m[i] for m in midi_features_all]), 0))

  midi_features = tuple(midi_features_all_concat)
  instrument_id = tf.concat(instrument_id_all, 0)

  midi_audio, midi_control_params, midi_synth_params = synthesis_generator. \
    gen_audio_from_cond_dict(conditioning_dict,
                             midi_features,
                             instrument_id=instrument_id,
                             display_progressbar=display_progressbar)
  return midi_audio, midi_control_params, midi_synth_params


def synthesize_mono_midi(synthesis_generator, expression_generator, midi_file,
                         instrument_id, output_dir,
                         pitch_offset=0,
                         speed_rate=1,
                         display_progressbar=True):
  """Synthesize monophonic MIDI file.
  If MIDI file contains more than one part, will synthesize the first part."""
  note_sequence = mono_midi_to_note_sequence(midi_file,
                                             tf.constant([instrument_id]),
                                             pitch_offset=pitch_offset,
                                             speed_rate=speed_rate)
  expression_generator_outputs = expression_generator(note_sequence, out=None,
                                                      training=False)
  conditioning_df = expression_generator_output_to_conditioning_df(
    expression_generator_outputs['output'], note_sequence)
  midi_audio, midi_control_params, midi_synth_params = conditioning_df_to_audio(
    synthesis_generator, conditioning_df,
    tf.constant([instrument_id]),
    display_progressbar=display_progressbar)
  if output_dir is not None:
    save_wav(midi_audio[0].numpy(), os.path.join(output_dir, os.path.basename(
      midi_file).replace('.mid', '.wav')),
             16000)
  return midi_audio, midi_control_params, midi_synth_params, conditioning_df


def fill_conditioning_df_with_mean(conditioning_df, note_expression_stat,
                                   instrument_abb):
  """Fill conditioning_df with mean note expression values.
  Used for ablation test."""
  note_expression_list = CONDITIONING_KEYS
  for note_expression in note_expression_list:
    value = np.ones_like(conditioning_df[note_expression].values) * \
            note_expression_stat[instrument_abb][
              note_expression]
    value[conditioning_df['pitch'].values == 0] = 0
    if note_expression == 'vibrato_extend':
      value[conditioning_df['note_length'].values < 50] = 0
    conditioning_df[note_expression] = value
  return conditioning_df


def conditioning_df_to_f0_ld(conditioning_df, instrument_id, ld):
  """Convert a conditioning_df to f0 and loudness contour.
  Used for ablation test."""
  q_pitch, _, _, onsets, offsets = conditioning_df_to_midi_features(
    conditioning_df)
  f0 = ddsp.core.midi_to_hz(q_pitch, midi_zero_silence=True)
  ld = np.ones_like(f0) * ld
  ld[q_pitch == 0] = -120
  data = {
    'f0_hz': f0,
    'loudness_db': ld,
    'instrument_id': tf.constant([instrument_id])
  }
  return data


def conditioning_df_to_midi(conditioning_df, programs_id):
  """Convert a conditioning_df to a PrettyMIDI object."""
  midi_data = pretty_midi.PrettyMIDI()
  instrument = pretty_midi.Instrument(program=programs_id)
  frame_ms = 4
  for _, note in conditioning_df.iterrows():
    pitch = int(note['pitch'])
    if pitch != 0:
      on = int(note['onset'])
      off = int(note['offset'])
      n = pretty_midi.Note(velocity=100,
                           pitch=int(note['pitch']),
                           start=on * frame_ms / 1000,
                           end=off * frame_ms / 1000)
      instrument.notes.append(n)
  midi_data.instruments.append(instrument)
  return midi_data



#  Copyright 2022 The MIDI-DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Decoders that predicts synthesis parameters from latent sequence z."""

from tqdm.autonotebook import tqdm
import tensorflow as tf
import ddsp
from ddsp.training import nn
from midi_ddsp.modules.cond_rnn import TwoLayerCondAutoregRNN

tfk = tf.keras
tfkl = tfk.layers

# Quantization constants for f0 and loudness.

# [-1, 1] in cents (0.01 per bin)
RELATIVE_F0_MIN = -1
RELATIVE_F0_MAX = 1
RELATIVE_F0_STEP = 0.01
ONE_OVER_RELATIVE_F0_STEP = int(round(1 / RELATIVE_F0_STEP))
RELATIVE_F0_BINS = 201

# [-120, 0], 1 per bin, "ld" means loudness
ABSOLUTE_LD_MIN = -120
ABSOLUTE_LD_MAX = 0
ABSOLUTE_LD_STEP = 1
ONE_OVER_ABSOLUTE_LD_STEP = int(round(1 / ABSOLUTE_LD_STEP))
ABSOLUTE_LD_BINS = 121

# [-1, 1] in cents (0.01 per bin)
ABSOLUTE_F0_MIN = 24
ABSOLUTE_F0_MAX = 96
ABSOLUTE_F0_STEP = 1 / 5
ONE_OVER_ABSOLUTE_F0_STEP = int(round(1 / ABSOLUTE_F0_STEP))
ABSOLUTE_F0_BINS = 361


def get_onehot_f0_dv(f0_dv):
  f0_dv = tf.clip_by_value(f0_dv[:, :, -1], RELATIVE_F0_MIN,
                           RELATIVE_F0_MAX) - RELATIVE_F0_MIN
  f0_dv = tf.cast(f0_dv / RELATIVE_F0_STEP, tf.int64)
  f0_dv_onehot = tf.one_hot(f0_dv, RELATIVE_F0_BINS)
  return f0_dv_onehot


def get_onehot_ld(ld):
  ld = tf.clip_by_value(ld[:, :, -1], ABSOLUTE_LD_MIN,
                        ABSOLUTE_LD_MAX) - ABSOLUTE_LD_MIN
  ld = tf.cast(ld / ABSOLUTE_LD_STEP, tf.int64)
  ld_onehot = tf.one_hot(ld, ABSOLUTE_LD_BINS)
  return ld_onehot


def get_float_f0_dv(f0_dv_onehot):
  f0_dv = tf.cast(tf.argmax(f0_dv_onehot, axis=-1) /
                  ONE_OVER_RELATIVE_F0_STEP + RELATIVE_F0_MIN,
                  tf.float32)
  return f0_dv[..., tf.newaxis]


def get_float_ld(ld_onehot):
  ld = tf.cast(tf.argmax(ld_onehot, axis=-1) /
               ONE_OVER_ABSOLUTE_LD_STEP + ABSOLUTE_LD_MIN,
               tf.float32)
  return ld[..., tf.newaxis]


def get_float_absolute_f0(f0_midi_onehot):
  f0_midi = tf.cast(tf.argmax(f0_midi_onehot, axis=-1) /
                    ONE_OVER_ABSOLUTE_F0_STEP + ABSOLUTE_LD_MIN,
                    tf.float32)
  return f0_midi[..., tf.newaxis]


def get_onehot_absolute_f0(f0_midi):
  f0_midi = tf.clip_by_value(f0_midi[:, :, -1], ABSOLUTE_F0_MIN,
                             ABSOLUTE_F0_MAX) - ABSOLUTE_F0_MIN
  f0_midi = tf.cast(f0_midi / ABSOLUTE_F0_STEP, tf.int64)
  f0_midi_onehot = tf.one_hot(f0_midi, ABSOLUTE_F0_BINS)
  return f0_midi_onehot


def top_p_sample(logits, p=0.95):
  logits_sort = tf.sort(logits, direction='DESCENDING', axis=-1)
  probs_sort = tf.nn.softmax(logits_sort)
  probs_sums = tf.cumsum(probs_sort, axis=1, exclusive=True)
  logits_masked = tf.where(probs_sums < p, logits_sort,
                           tf.ones_like(logits_sort) * 1000)
  min_logits = tf.reduce_min(input_tensor=logits_masked, axis=1, keepdims=True)
  logits = tf.where(
    logits < min_logits,
    tf.ones_like(logits, dtype=logits.dtype) * -1e10,
    logits,
  )
  return logits


def sample_from(logits, return_onehot=False, sample_method='top_p'):
  original_shape = logits.shape
  num_class = logits.shape[-1]
  if sample_method == 'argmax':
    value = tf.argmax(tf.reshape(logits, [-1, num_class]), axis=-1)
  elif sample_method == 'random':
    value = tf.random.categorical(tf.reshape(logits, [-1, num_class]), 1)
  elif sample_method == 'top_p':
    value = tf.random.categorical(
      top_p_sample(tf.reshape(logits, [-1, num_class])), 1)
  if return_onehot:
    value = tf.one_hot(value, num_class)
    return tf.reshape(value, original_shape)
  else:
    return tf.reshape(value, original_shape[:-1] + [1])


class MidiExpreToF0AutoregDecoder(TwoLayerCondAutoregRNN):
  """Autoregressive prediction of f0 from z_midi_decoder that contains MIDI and
   note expression controls."""

  def __init__(self, nhid=256, norm=True, dropout=0.5, sampling_method='top_p'):
    super().__init__(
      nhid=nhid,
      n_out=RELATIVE_F0_BINS,
      input_dropout=True,
      input_dropout_p=0.5,
      dropout=dropout,
    )
    self.birnn = tfkl.Bidirectional(tfkl.GRU(
      units=nhid, return_sequences=True, dropout=dropout
    ), )
    self.dense_out = tfkl.Dense(self.n_out)
    self.norm = nn.Normalize('layer') if norm else None
    self.sampling_method = sampling_method

  def split_teacher_force_output(self, output):
    f0_midi_dv_logits = output
    f0_midi_dv_onehot = get_onehot_f0_dv(get_float_f0_dv(f0_midi_dv_logits))
    output = {
      'f0_midi_dv_onehot': f0_midi_dv_onehot,
      'f0_midi_dv_logits': f0_midi_dv_logits,
    }
    return output

  def encode_cond(self, cond, training=False):
    cond = self.birnn(cond, training=training)
    return cond

  def decode_out(self, z_out, training=False):
    if self.norm is not None:
      z_out = self.norm(z_out)
    output = self.dense_out(z_out)
    return output

  def sample_out(self, out, training=False):
    f0_midi_dv_logits = out[..., :RELATIVE_F0_BINS]
    f0_midi_dv_onehot = sample_from(f0_midi_dv_logits,
                                    return_onehot=True,
                                    sample_method=self.sampling_method)
    output = {
      'f0_midi_dv_logits': f0_midi_dv_logits,
      'f0_midi_dv_onehot': f0_midi_dv_onehot,
      'curr_out': f0_midi_dv_onehot,
    }
    return output

  def preprocess(self, q_pitch, out):
    f0_midi = ddsp.core.hz_to_midi(out['f0_hz'])
    f0_midi_dv = f0_midi - q_pitch  # f0 residual
    f0_dv_onehot = get_onehot_f0_dv(f0_midi_dv)
    return f0_dv_onehot

  def postprocess(self, outputs, q_pitch, training=False):
    outputs['f0_midi_dv'] = get_float_f0_dv(outputs['f0_midi_dv_onehot'])
    outputs['f0_midi'] = outputs['f0_midi_dv'] + q_pitch
    outputs['f0_hz'] = ddsp.core.midi_to_hz(outputs['f0_midi'])
    return outputs

  def call(self, q_pitch, z_midi_decoder, conditioning_dict, out=None,
           training=False, display_progressbar=False):
    if training:
      out = self.preprocess(q_pitch, out)
      outputs = self.teacher_force(z_midi_decoder, out, training=training)
    else:
      outputs = self.autoregressive(z_midi_decoder, training=training,
                                    display_progressbar=display_progressbar)

    outputs = self.postprocess(outputs, q_pitch)

    return outputs


class MidiToF0LDAutoregDecoder(TwoLayerCondAutoregRNN):
  """Autoregressive prediction of f0 and loudnes from z_midi_decoder that
   only contains MIDI. (The implementation of MIDI2Params.)"""
  def __init__(self, nhid=256, norm=True, dropout=0.5, sampling_method='top_p'):
    super().__init__(
      nhid=nhid,
      n_out=RELATIVE_F0_BINS + ABSOLUTE_LD_BINS,
      input_dropout=True,
      input_dropout_p=0.5,
      dropout=dropout,
    )
    self.birnn = tfkl.Bidirectional(tfkl.GRU(
      units=nhid, return_sequences=True, dropout=dropout
    ), )
    self.dense_out = nn.FcStackOut(ch=nhid, layers=2, n_out=self.n_out)
    self.dense_in = nn.FcStack(ch=nhid, layers=2)
    self.norm = nn.Normalize('layer') if norm else None
    self.sampling_method = sampling_method

  def encode_z(self, z, training=False):
    return self.dense_in(z)

  def split_teacher_force_output(self, output):
    f0_midi_dv_logits = output[..., :RELATIVE_F0_BINS]
    f0_midi_dv_onehot = get_onehot_f0_dv(get_float_f0_dv(f0_midi_dv_logits))
    ld_logits = output[...,
                RELATIVE_F0_BINS:RELATIVE_F0_BINS + ABSOLUTE_LD_BINS]
    ld_onehot = get_onehot_ld(get_float_ld(ld_logits))
    output = {
      'f0_midi_dv_logits': f0_midi_dv_logits,
      'f0_midi_dv_onehot': f0_midi_dv_onehot,
      'ld_logits': ld_logits,
      'ld_onehot': ld_onehot,
    }
    return output

  def encode_cond(self, cond, training=False):
    cond = self.birnn(cond, training=training)
    return cond

  def decode_out(self, z_out, training=False):
    if self.norm is not None:
      z_out = self.norm(z_out)
    output = self.dense_out(z_out)
    return output

  def sample_out(self, out, training=False):
    f0_midi_dv_logits = out[..., :RELATIVE_F0_BINS]
    f0_midi_dv_onehot = sample_from(f0_midi_dv_logits, return_onehot=True,
                                    sample_method=self.sampling_method)
    ld_logits = out[..., RELATIVE_F0_BINS:RELATIVE_F0_BINS + ABSOLUTE_LD_BINS]
    ld_onehot = sample_from(ld_logits, return_onehot=True,
                            sample_method=self.sampling_method)
    output = {
      'f0_midi_dv_logits': f0_midi_dv_logits,
      'f0_midi_dv_onehot': f0_midi_dv_onehot,
      'ld_logits': ld_logits,
      'ld_onehot': ld_onehot,
      'curr_out': tf.concat([f0_midi_dv_onehot, ld_onehot], -1)
    }
    return output

  def preprocess(self, q_pitch, out):
    f0_midi = ddsp.core.hz_to_midi(out['f0_hz'])
    f0_midi_dv = f0_midi - q_pitch  # f0 residual
    f0_midi_dv_onehot = get_onehot_f0_dv(f0_midi_dv)
    ld = out['loudness_db']
    ld_onehot = get_onehot_ld(ld)
    return tf.concat([f0_midi_dv_onehot, ld_onehot], -1)

  def postprocess(self, outputs, q_pitch, training=False):
    outputs['f0_midi_dv'] = get_float_f0_dv(outputs['f0_midi_dv_onehot'])
    outputs['f0_midi'] = outputs['f0_midi_dv'] + q_pitch
    outputs['f0_hz'] = ddsp.core.midi_to_hz(outputs['f0_midi'])
    outputs['ld'] = get_float_ld(outputs['ld_onehot'])
    return outputs

  def call(self, q_pitch, z_midi_decoder, out=None, training=False):

    if training:
      out = self.preprocess(q_pitch, out)
      outputs = self.teacher_force(z_midi_decoder, out, training=training)
    else:
      outputs = self.autoregressive(z_midi_decoder, training=training)

    outputs = self.postprocess(outputs, q_pitch)

    return outputs

  def partial_teacher_force(self, q_pitch, z_midi_decoder, out,
                            teacher_force_mask, training=False,
                            display_progressbar=False):
    """For the expression generator reconstruction comparison.
    (Table 1 (c) in paper)"""
    # right shift and encode out
    out = self.preprocess(q_pitch, out)
    go_frame = tf.tile([[0.0]], [out.shape[0], out.shape[-1]])[:, tf.newaxis, :]
    out = tf.concat([go_frame, out[:, :-1, :]], 1)
    out = self.encode_out(out)

    cond = self.encode_cond(z_midi_decoder, training=training)
    batch_size = cond.shape[0]
    length = cond.shape[1]
    prev_out = tf.tile([[0.0]], [batch_size, self.n_out])[:, tf.newaxis,
               :]  # go_frame
    prev_states = (None, None)
    overall_outputs = []

    teacher_force_batch_size = teacher_force_mask.shape[0]

    for i in tqdm(range(length), position=0, leave=True, desc='Generating: ',
                  disable=not display_progressbar):
      curr_cond = tf.tile(cond[:, i:i + 1, :], [teacher_force_batch_size, 1, 1])
      prev_out = self.encode_out(prev_out) * (
            1 - teacher_force_mask[:, i:i + 1, :]) + \
                 out[:, i:i + 1, :] * teacher_force_mask[:, i:i + 1, :]
      curr_out, curr_states = self._one_step(curr_cond, prev_out, prev_states,
                                             training=training)
      curr_out = self.sample_out(curr_out)
      overall_outputs.append(curr_out)
      prev_out, prev_states = curr_out['curr_out'], curr_states

    outputs = {}
    for k in curr_out:
      outputs[k] = tf.concat([x[k] for x in overall_outputs], 1)

    return self.postprocess(outputs, q_pitch)


class F0LDAutoregDecoder(TwoLayerCondAutoregRNN):
  """Autoregressive prediction of f0 and loudness.
  (A fully unconditional model without MIDI input.)"""
  def __init__(self, nhid=256, norm=True, dropout=0.5, sampling_method='top_p'):
    self.n_f0 = ABSOLUTE_F0_BINS
    self.n_ld = ABSOLUTE_LD_BINS
    super().__init__(
      nhid=nhid,
      n_out=self.n_f0 + self.n_ld,
      input_dropout=True,
      input_dropout_p=0.5,
      dropout=dropout,
    )
    self.birnn = tfkl.Bidirectional(tfkl.GRU(
      units=nhid, return_sequences=True, dropout=dropout
    ), )
    self.dense_out = nn.FcStackOut(ch=nhid, layers=2, n_out=self.n_out)
    self.dense_in = nn.FcStack(ch=nhid, layers=2)
    self.norm = nn.Normalize('layer') if norm else None
    self.sampling_method = sampling_method

  def split_teacher_force_output(self, output):
    # split outputs
    f0_midi_logits = output[..., :self.n_f0]
    f0_midi_onehot = get_onehot_absolute_f0(
      get_float_absolute_f0(f0_midi_logits))
    ld_logits = output[..., self.n_f0:self.n_f0 + self.n_ld]
    ld_onehot = get_onehot_ld(get_float_ld(ld_logits))
    output = {
      'f0_midi_logits': f0_midi_logits,
      'f0_midi_onehot': f0_midi_onehot,
      'ld_logits': ld_logits,
      'ld_onehot': ld_onehot,
    }
    return output

  def encode_z(self, z, training=False):
    return self.dense_in(z)

  def decode_out(self, z_out, training=False):
    if self.norm is not None:
      z_out = self.norm(z_out)
    output = self.dense_out(z_out)
    return output

  def sample_out(self, out, training=False):
    f0_midi_logits = out[..., :self.n_f0]
    f0_midi_onehot = sample_from(f0_midi_logits, return_onehot=True,
                                 sample_method=self.sampling_method)
    ld_logits = out[..., self.n_f0:self.n_f0 + self.n_ld]
    ld_onehot = sample_from(ld_logits, return_onehot=True,
                            sample_method=self.sampling_method)
    output = {
      'f0_midi_logits': f0_midi_logits,
      'f0_midi_onehot': f0_midi_onehot,
      'ld_logits': ld_logits,
      'ld_onehot': ld_onehot,
      'curr_out': tf.concat([f0_midi_onehot, ld_onehot], -1)
    }
    return output

  def preprocess(self, out, training=False):
    f0_midi = ddsp.core.hz_to_midi(out['f0_hz'])
    f0_midi_onehot = get_onehot_absolute_f0(f0_midi)
    ld = out['loudness_db']
    ld_onehot = get_onehot_ld(ld)
    return tf.concat([f0_midi_onehot, ld_onehot], -1)

  def postprocess(self, outputs, training=False):
    outputs['f0_midi'] = get_float_absolute_f0(outputs['f0_midi_onehot'])
    outputs['f0_hz'] = ddsp.core.midi_to_hz(outputs['f0_midi'])
    outputs['ld'] = get_float_ld(outputs['ld_onehot'])
    return outputs

  def call(self, q_pitch, z_midi_decoder, out=None, training=False):
    z_midi_decoder = z_midi_decoder[..., 66:]

    if training:
      out = self.preprocess(out)
      outputs = self.teacher_force(z_midi_decoder, out, training=training)
    else:
      outputs = self.autoregressive(z_midi_decoder, training=training)

    outputs = self.postprocess(outputs)

    return outputs


class MidiExpreToSynthDecoder(nn.DictLayer):  # TODO: (yusongwu) merge into DDSP
  """Predicting synthesis parameters from MIDI and note expression controls
   using only dilated convolution networks."""
  def __init__(self,
               net=None,
               norm=True,
               output_splits=(('f0_midi', 1),
                              ('amplitudes', 1),
                              ('harmonic_distribution', 60),
                              ('magnitudes', 65)),
               **kwargs):
    """Constructor."""
    self.output_splits = output_splits
    self.n_out = sum([v[1] for v in output_splits])
    output_keys = [v[0] for v in output_splits]
    super().__init__(output_keys=output_keys, **kwargs)

    # Layers.
    self.net = net
    self.dense_out = tfkl.Dense(self.n_out)
    self.norm = nn.Normalize('layer') if norm else None

  def call(self, z, training=False):
    x = self.net(z, training=training)

    if self.norm is not None:
      x = self.norm(x)

    x = self.dense_out(x)

    outputs = nn.split_to_dict(x, self.output_splits)
    return outputs


class MidiToSynthAutoregDecoder(tfkl.Layer):
  """Predicting synthesis parameters from MIDI and note expression controls
   using autoregressive RNN for f0 and dilated convolution networks for
   amplitudes, harmonic distribution and noise magnitudes."""
  def __init__(self, nhid=256,
               hd_noise_output_splits=(('amplitudes', 1),
                                       ('harmonic_distribution', 60),
                                       ('noise_magnitudes', 65))):
    super().__init__()
    self.q_pitch_emb = tfkl.Dense(64)
    self.midi_to_f0 = MidiExpreToF0AutoregDecoder(nhid=nhid)
    self.midi_f0_to_harmonic = MidiExpreToSynthDecoder(
      net=nn.DilatedConvStack(
        ch=128,
        layers_per_stack=5,
        stacks=4,
        norm_type='layer',
        conditional=False,
      ),
      output_splits=hd_noise_output_splits,
    )

  def call(self, q_pitch, z_midi_decoder, conditioning_dict, out=None,
           training=False, display_progressbar=False):
    f0_output = self.midi_to_f0(q_pitch, z_midi_decoder, conditioning_dict,
                                out=out, training=training,
                                display_progressbar=display_progressbar)

    if training:
      f0_midi = ddsp.core.hz_to_midi(out['f0_hz'])
    else:
      f0_midi = f0_output['f0_midi']

    z = tf.concat([z_midi_decoder, self.q_pitch_emb(f0_midi / 127)], -1)
    outputs = self.midi_f0_to_harmonic(z, training=training)

    outputs['f0_output'] = f0_output
    if training:
      outputs['f0_hz'] = out['f0_hz']
    else:
      outputs['f0_hz'] = ddsp.core.midi_to_hz(f0_output['f0_midi'],
                                             midi_zero_silence=True)

    return outputs


class MidiNoiseToHarmonicDecoder(nn.DictLayer):
  """Decodes MIDI notes (& velocities) to f0, amps, hd, noise."""

  def __init__(self,
               net=None,
               f0_residual=True,
               norm=True,
               output_splits=(('f0_midi', 1),
                              ('amplitudes', 1),
                              ('harmonic_distribution', 60),
                              ('magnitudes', 65)),
               **kwargs):
    """Constructor."""
    self.output_splits = output_splits
    self.n_out = sum([v[1] for v in output_splits])
    output_keys = [v[0] for v in output_splits] + ['f0_hz']
    super().__init__(output_keys=output_keys, **kwargs)

    # Layers.
    self.net = net
    self.f0_residual = f0_residual
    self.dense_out = tfkl.Dense(self.n_out)
    self.norm = nn.Normalize('layer') if norm else None

  def call(self, noise, z_pitch, z=None):
    """Forward pass for the MIDI decoder.

    Args:
      z_pitch: Tensor containing encoded pitch in MIDI scale. [batch, time, 1].
      z_vel: Tensor containing encoded velocity in MIDI scale. [batch, time, 1].
      z: Additional non-MIDI latent tensor. [batch, time, n_z]

    Returns:
      A dictionary to feed into a processor group.
    """
    x = noise
    x = self.net(x) if z is None else self.net([x, z])

    if self.norm is not None:
      x = self.norm(x)

    x = self.dense_out(x)

    outputs = nn.split_to_dict(x, self.output_splits)

    if self.f0_residual:
      outputs['f0_midi'] += z_pitch

    outputs['f0_hz'] = ddsp.core.midi_to_hz(outputs['f0_midi'])
    return outputs



#  Copyright 2022 The MIDI-DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""LossHelper for handling the reconstruction used in training."""

import tensorflow as tf
import ddsp
from midi_ddsp.modules.loss_helper import LossHelper
from ddsp.losses import ParamLoss

tfk = tf.keras
tfkl = tfk.layers


class EvalReconLossGroup(tfkl.Layer):
  """Evaluation loss group, for calculating L1 and L2 loss
  on synthesis parameters and spectral loss on audio.
  It is used in evaluation phase regardless of the training objective."""

  def __init__(self):
    super().__init__()
    self.f0_loss = ParamLoss(weight=50.0, loss_type='L2',
                             name='f0_reconstruction')
    self.amps_loss = ParamLoss(weight=0.5, loss_type='L1',
                               name='amplitude_reconstruction')
    self.hd_loss = ParamLoss(weight=500.0, loss_type='L1',
                             name='harmonic_distribution_reconstruction')
    self.noise_loss = ParamLoss(weight=0.5, loss_type='L1',
                                name='noise_reconstruction')
    self.synth_spectral_loss = ddsp.losses.SpectralLoss(loss_type='L1',
                                                        mag_weight=1.0,
                                                        logmag_weight=1.0)
    self.midi_spectral_loss = ddsp.losses.SpectralLoss(loss_type='L1',
                                                       mag_weight=1.0,
                                                       logmag_weight=1.0)

  def call(self, inputs, outputs, synth_coder_only=False):
    loss_spectral_synth = self.synth_spectral_loss(inputs['audio'],
                                                   outputs['synth_audio'])
    if synth_coder_only:
      loss_dict = {
        'loss_spectral_synth': loss_spectral_synth,
        'total_loss': loss_spectral_synth
      }
      return loss_dict

    loss_f0 = self.f0_loss(outputs['params_pred']['f0_output']['f0_midi'],
                           ddsp.core.hz_to_midi(inputs['f0_hz']),
                           weights=outputs['f0_loss_weights'])
    loss_amp = self.amps_loss(outputs['amps_pred'], outputs['amps'])
    loss_hd = self.hd_loss(outputs['hd_pred'], outputs['hd'])
    loss_noise = self.noise_loss(outputs['noise_pred'], outputs['noise'])
    loss_spectral_midi = self.midi_spectral_loss(inputs['audio'],
                                                 outputs['midi_audio'])
    loss_dict = {
      'loss_amp': loss_amp,
      'loss_f0': loss_f0,
      'loss_hd': loss_hd,
      'loss_noise': loss_noise,
      'loss_spectral_synth': loss_spectral_synth,
      'loss_spectral_midi': loss_spectral_midi,
      'total_loss': loss_f0 + loss_amp + loss_hd + loss_noise
    }

    return loss_dict


class ReconLossGroup(tfkl.Layer):
  """The reconstruction loss group used in MIDI-DDSP.
  It calculates back-prop (in total_loss) consists of spectral loss on audio,
  L1 and L2 loss on amplitude, harmonic distribution and noise magnitude, and
  cross-entropy loss on f0.
  However, other losses are also calculated but not back-proped to monitor
  the training process."""

  def __init__(self, midi_audio_loss=False, synth_params_loss=True):
    super().__init__()
    self.amps_loss = ParamLoss(weight=0.5, loss_type='L1',
                               name='amplitude_reconstruction')
    self.hd_loss = ParamLoss(weight=500.0, loss_type='L1',
                             name='harmonic_distribution_reconstruction')
    self.noise_loss = ParamLoss(weight=0.5, loss_type='L1',
                                name='noise_reconstruction')
    self.synth_spectral_loss = ddsp.losses.SpectralLoss(loss_type='L1',
                                                        mag_weight=1.0,
                                                        logmag_weight=1.0)
    self.midi_audio_loss = midi_audio_loss
    self.synth_params_loss = synth_params_loss
    self.midi_spectral_loss = ddsp.losses.SpectralLoss(loss_type='L1',
                                                       mag_weight=1.0,
                                                       logmag_weight=1.0)

  def call(self, inputs, outputs, synth_coder_only=False, add_synth_loss=False):
    loss_spectral_synth = self.synth_spectral_loss(inputs['audio'],
                                                   outputs['synth_audio'])
    if synth_coder_only:
      loss_dict = {
        'loss_spectral_synth': loss_spectral_synth,
        'total_loss': loss_spectral_synth
      }
      return loss_dict

    f0_dv = ddsp.core.hz_to_midi(inputs['f0_hz']) - tf.stop_gradient(
      outputs['q_pitch'])
    f0_midi_bined = tf.cast(
      (tf.clip_by_value(f0_dv[:, :, -1], -1, 1) + 1) * 100, tf.int64)
    loss_f0 = tf.reduce_mean(
      tf.nn.sparse_softmax_cross_entropy_with_logits(f0_midi_bined,
                                                     outputs['params_pred'][
                                                       'f0_output'][
                                                       'f0_midi_dv_logits']))

    loss_amp = self.amps_loss(outputs['amps_pred'], outputs['amps'])
    loss_hd = self.hd_loss(outputs['hd_pred'], outputs['hd'])
    loss_noise = self.noise_loss(outputs['noise_pred'], outputs['noise'])

    loss_spectral_midi = self.midi_spectral_loss(inputs['audio'],
                                                 outputs['midi_audio'])
    loss_dict = {
      'loss_amp': loss_amp,
      'loss_f0': loss_f0,
      'loss_hd': loss_hd,
      'loss_noise': loss_noise,
      'loss_spectral_synth': loss_spectral_synth,
      'loss_spectral_midi': loss_spectral_midi,
    }
    if self.synth_params_loss:
      loss_dict['total_loss'] = loss_amp + loss_f0 + loss_hd + loss_noise
    else:
      loss_dict['total_loss'] = loss_f0
    if add_synth_loss:
      loss_dict['total_loss'] = loss_dict['total_loss'] + loss_dict[
        'loss_spectral_synth']
    if self.midi_audio_loss:
      loss_dict['total_loss'] = loss_dict['total_loss'] + loss_dict[
        'loss_spectral_midi']

    return loss_dict


class F0LdReconLossGroup(tfkl.Layer):
  """Reconstruction loss group for MIDI2Params.
  Basically calculate cross-entropy loss on f0 and loundess logits.
  However, L1 and L2 loss on synthesis parameters and spectral loss on audio
  are also calculated but not back-proped to monitor the training process. """

  def __init__(self, midi_audio_loss=False, synth_params_loss=True,
               full_f0_loss=False):
    super().__init__()
    self.hd_loss = ParamLoss(weight=500.0, loss_type='L1',
                             name='harmonic_distribution_reconstruction')
    self.noise_loss = ParamLoss(weight=0.5, loss_type='L1',
                                name='noise_reconstruction')
    self.synth_spectral_loss = ddsp.losses.SpectralLoss(loss_type='L1',
                                                        mag_weight=1.0,
                                                        logmag_weight=1.0)
    self.midi_audio_loss = midi_audio_loss
    self.synth_params_loss = synth_params_loss
    self.midi_spectral_loss = ddsp.losses.SpectralLoss(loss_type='L1',
                                                       mag_weight=1.0,
                                                       logmag_weight=1.0)
    self.full_f0_loss = full_f0_loss

  def call(self, inputs, outputs, synth_coder_only=False, add_synth_loss=False):
    loss_spectral_synth = self.synth_spectral_loss(inputs['audio'],
                                                   outputs['synth_audio'])
    if synth_coder_only:
      loss_dict = {
        'loss_spectral_synth': loss_spectral_synth,
        'total_loss': loss_spectral_synth
      }
      return loss_dict

    amps_bined = tf.cast(
      tf.clip_by_value(tf.stop_gradient(inputs['loudness_db'] + 120)[:, :, -1],
                       0, 120),
      tf.int64)
    if self.full_f0_loss:
      f0_midi = ddsp.core.hz_to_midi(inputs['f0_hz'])
      f0_midi_bined = tf.cast(
        (tf.clip_by_value(f0_midi[:, :, -1], 24, 96) - 24) * 5, tf.int64)
      loss_f0 = tf.reduce_mean(
        tf.nn.sparse_softmax_cross_entropy_with_logits(f0_midi_bined,
                                                       outputs['params_pred'][
                                                         'f0_midi_logits']))

    else:
      f0_dv = ddsp.core.hz_to_midi(inputs['f0_hz']) - tf.stop_gradient(
        outputs['q_pitch'])
      f0_midi_bined = tf.cast(
        (tf.clip_by_value(f0_dv[:, :, -1], -1, 1) + 1) * 100, tf.int64)
      loss_f0 = tf.reduce_mean(
        tf.nn.sparse_softmax_cross_entropy_with_logits(f0_midi_bined,
                                                       outputs['params_pred'][
                                                         'f0_midi_dv_logits']))
    loss_amp = tf.reduce_mean(
      tf.nn.sparse_softmax_cross_entropy_with_logits(amps_bined,
                                                     outputs['params_pred'][
                                                       'ld_logits']))
    loss_hd = self.hd_loss(outputs['hd_pred'], outputs['hd'])
    loss_noise = self.noise_loss(outputs['noise_pred'], outputs['noise'])
    loss_spectral_midi = self.midi_spectral_loss(inputs['audio'],
                                                 outputs['midi_audio'])
    loss_dict = {
      'loss_amp': loss_amp,
      'loss_f0': loss_f0,
      'loss_hd': loss_hd,
      'loss_noise': loss_noise,
      'loss_spectral_synth': loss_spectral_synth,
      'loss_spectral_midi': loss_spectral_midi,
      'total_loss': loss_amp + loss_f0,
    }
    if add_synth_loss:
      loss_dict['total_loss'] = loss_dict['total_loss'] + loss_dict[
        'loss_spectral_synth']

    return loss_dict


class ReconLossHelper(LossHelper):
  """Overall wrapper for getting the reconstruction loss helper
  based on hyperparameters."""

  def __init__(self, hp, eval_recon_loss=False):
    midi_audio_loss = hp.midi_audio_loss
    synth_params_loss = hp.synth_params_loss
    midi_decoder_type = hp.midi_decoder_type
    midi_decoder_decoder_net = hp.midi_decoder_decoder_net
    super().__init__()

    self.loss_list = [
      'loss_amp',
      'loss_f0',
      'loss_hd',
      'loss_noise',
      'loss_spectral_synth',
      'loss_spectral_midi',
      'total_loss'
    ]

    if eval_recon_loss:
      self.loss_group = EvalReconLossGroup()

    elif midi_decoder_type == 'midi_decoder' or \
          midi_decoder_type == 'unconditioned' or \
          midi_decoder_decoder_net == 'rnn_f0_ld':

      self.loss_group = F0LdReconLossGroup(midi_audio_loss=midi_audio_loss,
                                           synth_params_loss=synth_params_loss,
                                           full_f0_loss=
                                           midi_decoder_type == 'unconditioned')
    elif midi_decoder_type == 'interpretable_conditioning':
      if 'dilated_conv' in midi_decoder_decoder_net:
        self.loss_group = EvalReconLossGroup()
      else:
        self.loss_group = ReconLossGroup(midi_audio_loss=midi_audio_loss,
                                         synth_params_loss=synth_params_loss)

    self.init_metrics()



#  Copyright 2022 The MIDI-DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Synthesize any MIDI file using MIDI-DDSP through command line."""

# Ignore a bunch of warnings
import os

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

import tensorflow as tf
import numpy as np
import pretty_midi
import argparse
import glob
from tqdm.autonotebook import tqdm

from midi_ddsp.data_handling.instrument_name_utils import \
  INST_NAME_TO_MIDI_PROGRAM_DICT, MIDI_PROGRAM_TO_INST_ID_DICT, \
  MIDI_PROGRAM_TO_INST_NAME_DICT
from midi_ddsp.utils.midi_synthesis_utils import note_list_to_sequence, \
  expression_generator_output_to_conditioning_df, batch_conditioning_df_to_audio
from midi_ddsp.utils.audio_io import save_wav
from midi_ddsp.utils.training_utils import get_hp
from midi_ddsp.utils.inference_utils import ensure_same_length
from midi_ddsp.utils.file_utils import pickle_dump
from midi_ddsp.hparams_synthesis_generator import hparams as hp
from midi_ddsp.modules.get_synthesis_generator import get_synthesis_generator, \
  get_fake_data_synthesis_generator
from midi_ddsp.modules.expression_generator import ExpressionGenerator, \
  get_fake_data_expression_generator

FRAME_RATE = 250
SAMPLE_RATE = 16000


def load_pretrained_model(synthesis_generator_path=None,
                          expression_generator_path=None):
  """Load pre-trained model weights."""
  package_dir = os.path.dirname(os.path.realpath(__file__))

  if not os.path.exists(os.path.join(package_dir,
                                     'midi_ddsp_model_weights_urmp_9_10')):
    raise FileNotFoundError('Model weights not found. '
                            'Please run \'midi_ddsp_download_model_weights\' '
                            'to download model weights, '
                            'or specify path to model weights.')

  if synthesis_generator_path is None:
    synthesis_generator_path = os.path.join(
      package_dir,
      'midi_ddsp_model_weights_urmp_9_10',
      'synthesis_generator',
      '50000')
  if expression_generator_path is None:
    expression_generator_path = os.path.join(
      package_dir,
      'midi_ddsp_model_weights_urmp_9_10',
      'expression_generator',
      '5000')

  hp_dict = get_hp(
    os.path.join(os.path.dirname(synthesis_generator_path), 'train.log'))
  for k, v in hp_dict.items():
    setattr(hp, k, v)
  synthesis_generator = get_synthesis_generator(hp)
  synthesis_generator._build(get_fake_data_synthesis_generator(hp))
  synthesis_generator.load_weights(synthesis_generator_path).expect_partial()

  n_out = 6
  expression_generator = ExpressionGenerator(n_out=n_out, nhid=128)
  fake_data = get_fake_data_expression_generator(n_out)
  _ = expression_generator(fake_data['cond'], out=fake_data['target'],
                           training=True)
  expression_generator.load_weights(expression_generator_path).expect_partial()
  return synthesis_generator, expression_generator


def synthesize_midi(synthesis_generator, expression_generator, midi_file,
                    pitch_offset=0, speed_rate=1.0,
                    output_dir=None,
                    use_fluidsynth=False,
                    sf2_path='/usr/share/sounds/sf2/FluidR3_GM.sf2',
                    display_progressbar=True,
                    skip_existing_files=False,
                    save_metadata=False):
  """
  Synthesize a midi file using MIDI-DDSP.
  Args:
      synthesis_generator: The instance of a Synthesis Generator.
      expression_generator: The instance of a expression generator.
      midi_file: The path to the MIDI file.
      pitch_offset: Pitch in semitone to transpose.
      speed_rate: The speed to synthesize the MIDI file.
      output_dir: The directory for saving outputs.
        If output_dir is None, outputs will not be written to disk.
      use_fluidsynth: Whether to use FluidSynth for synthesizing instruments
        that are not available in MIDI-DDSP.
      sf2_path: The path to a sf2 soundfont file used for FluidSynth.
      display_progressbar: Whether to display progress bar.
      skip_existing_files: Skip synthesizing MIDI files if already exist
        output folders.
      save_metadata: also save metadata containing instrument_id, note
        expression controls, and synthesis parameters.

  Returns: A dict of output:
          'mix_audio': mix audio,
          'stem_audio': stem audios,
          'part_synth_by_model': list of part numbers that are synthesized
            by MIDI-DDSP,
          'midi_control_params': control parameters generated by MIDI-DDSP,
          'midi_synth_params': synth parameters generated by MIDI-DDSP,
          'conditioning_df': note expressions predicted by expression generator
            in the format of DataFrame,
  """
  # Check if there is existing files.
  filename = os.path.splitext(os.path.basename(midi_file))[0]
  if output_dir is not None:
    output_dir = os.path.join(output_dir, filename)
    if os.path.exists(output_dir) and skip_existing_files:
      print(f'{midi_file} has been synthesized, will skip this file.')
      return

  # Get all the midi program in URMP dataset excluding guitar.
  allowed_midi_program = list(INST_NAME_TO_MIDI_PROGRAM_DICT.values())[:-1]
  midi_data = pretty_midi.PrettyMIDI(midi_file)
  instrument_id_all = []
  conditioning_df_all = []
  part_synth_by_model = []
  midi_audio_all = {}
  midi_synth_params_all = {}
  midi_control_params_all = {}

  # For each part, predict expressions using MIDI-DDSP,
  # or synthesize using FluidSynth.
  for part_number, instrument in enumerate(midi_data.instruments):
    midi_program = instrument.program
    if midi_program in allowed_midi_program:
      note_sequence = note_list_to_sequence(instrument.notes, fs=FRAME_RATE,
                                            pitch_offset=pitch_offset,
                                            speed_rate=speed_rate)
      instrument_id = tf.constant([MIDI_PROGRAM_TO_INST_ID_DICT[midi_program]])
      instrument_id_all.append(instrument_id)
      note_sequence['instrument_id'] = instrument_id
      expression_generator_outputs = expression_generator(note_sequence,
                                                          out=None,
                                                          training=False)
      conditioning_df = expression_generator_output_to_conditioning_df(
        expression_generator_outputs['output'], note_sequence)
      conditioning_df_all.append(conditioning_df)
      part_synth_by_model.append(part_number)
    elif use_fluidsynth:
      instrument_name = MIDI_PROGRAM_TO_INST_NAME_DICT[midi_program]
      print(
        f'Part {part_number} in {midi_file} has {instrument_name} as '
        f'instrument which cannot be synthesized by model. '
        f'Using fluidsynth instead.')

      fluidsynth_wav_r3 = instrument.fluidsynth(SAMPLE_RATE, sf2_path=sf2_path)
      fluidsynth_wav_r3 *= 0.25  # * 0.25 for lower volume
      midi_audio_all[part_number] = fluidsynth_wav_r3

  # Synthesize audio in batch using Synthesis Generator.
  if len(conditioning_df_all) > 0:
    midi_audio, midi_control_params, midi_synth_params = \
      batch_conditioning_df_to_audio(
        synthesis_generator,
        conditioning_df_all,
        instrument_id_all,
        display_progressbar=display_progressbar)

    midi_synth_params = midi_synth_params['inputs']
    conditioning_df_all_for_save = {}
    instrument_id_all_for_save = {}
    for i in range(midi_audio.shape[0]):
      part_number = part_synth_by_model[i]

      # align audio with part number
      midi_audio_all[part_number] = midi_audio[i].numpy()

      # align instrument_id with part number
      instrument_id_all_for_save[part_number] = instrument_id_all[i].numpy()[0]

      # align note expression controls with part number
      conditioning_df_all_for_save[part_number] = conditioning_df_all[i]

      # align synthesis parameters with part number
      # get the midi synth parameters
      midi_synth_params_all[part_number] = {
        k: v[i].numpy() for k, v in midi_synth_params.items()}

      # align control parameters with part number
      # (yusongwu) sorry for mis-aligned variable names between
      # synth_params and control_params due to historical issues
      midi_control_params_all[part_number] = {
        'amplitudes': midi_control_params[1][i].numpy(),
        'harmonic_distribution': midi_control_params[2][i].numpy(),
        'noise_magnitudes': midi_control_params[3][i].numpy(),
        'f0_hz': midi_control_params[0][i].numpy(),
      }

  # Sorting out and save the wav.
  if output_dir is not None:
    os.makedirs(output_dir, exist_ok=True)
    for part_number, instrument in enumerate(midi_data.instruments):
      midi_program = instrument.program
      instrument_name = MIDI_PROGRAM_TO_INST_NAME_DICT[midi_program]
      if midi_program in allowed_midi_program:
        audio = midi_audio_all[part_number]
        save_wav(audio,
                 os.path.join(
                   output_dir,
                   f'{part_number}_{instrument_name}.wav'),
                 SAMPLE_RATE)
      elif use_fluidsynth:
        audio = midi_audio_all[part_number]
        save_wav(audio,
                 os.path.join(
                   output_dir,
                   f'{part_number}_{instrument_name}_fluidsynth.wav'),
                 SAMPLE_RATE)

  # If there is audio synthesized, mix the audio and return the output.
  if midi_audio_all:
    midi_audio_mix = np.sum(
      np.stack(ensure_same_length(
        [a.astype(np.float64) for a in midi_audio_all.values()], axis=0),
        axis=-1),
      axis=-1)
    if output_dir is not None:
      save_wav(midi_audio_mix, os.path.join(output_dir, 'mix.wav'), SAMPLE_RATE)
      if save_metadata:
        metadata = {
            'instrument_id': instrument_id_all_for_save,
            'note_expression_control': conditioning_df_all_for_save,
            'synthesis_parameters': midi_synth_params_all,
        }
        pickle_dump(metadata, os.path.join(output_dir, 'metadata.pickle'))
    output = {
      'mix_audio': midi_audio_mix,
      'stem_audio': midi_audio_all,
      'part_synth_by_model': part_synth_by_model,
      'midi_control_params': midi_control_params_all,
      'midi_synth_params': midi_synth_params_all,
      'conditioning_df': conditioning_df_all,
    }
  else:
    output = None

  return output


def main():
  parser = argparse.ArgumentParser(
    description='Synthesize MIDI files using MIDI-DDSP.')
  parser.add_argument('--midi_dir', type=str, default=None,
                      help='The directory containing MIDI files.')
  parser.add_argument('--midi_path', type=str, default=None,
                      help='The path to a MIDI file.')
  parser.add_argument('--pitch_offset', type=int, default=0,
                      help='Pitch offset to transpose in semitone.')
  parser.add_argument('--speed_rate', type=float, default=1.0,
                      help='The speed to synthesize the MIDI file(s).')
  parser.add_argument('--sf2_path', type=str,
                      default=None,
                      help='The path to a sf2 soundfont file.')
  parser.add_argument('--output_dir', type=str, default=None,
                      help='The directory for output audio.')
  parser.add_argument('--use_fluidsynth', action='store_true',
                      help='Use FluidSynth to synthesize the midi instruments '
                           'that are not contained in MIDI-DDSP.')
  parser.add_argument('--synthesis_generator_weight_path', type=str,
                      default=None,
                      help='The path to the Synthesis Generator weights. '
                           'It is not a specific file path but an index path. '
                           'See https://www.tensorflow.org/guide/checkpoint#'
                           'restore_and_continue_training.')
  parser.add_argument('--expression_generator_weight_path', type=str,
                      default=None,
                      help='The path to the expression generator weights. '
                           'It is not a specific file path but an index path. '
                           'See https://www.tensorflow.org/guide/checkpoint#'
                           'restore_and_continue_training.')
  parser.add_argument('--skip_existing_files', dest='skip_existing_files',
                      action='store_true',
                      help='Skip synthesizing MIDI files if already exist '
                           'output folders.')
  parser.add_argument('--save_metadata', dest='save_metadata',
                      action='store_true',
                      help='Save metadata including containing instrument_id, '
                           'note expression controls, and synthesis '
                           'parameters generated by MIDI-DDSP.')

  args = parser.parse_args()

  synthesis_generator, expression_generator = load_pretrained_model(
    synthesis_generator_path=args.synthesis_generator_weight_path,
    expression_generator_path=args.expression_generator_weight_path)

  if args.output_dir is None:
    print('Output directory not specified. Output to current directory.')
    output_dir = './'
  else:
    output_dir = args.output_dir

  if args.midi_dir and args.midi_path:
    print('Both midi_dir and midi_path are provided. Will use midi_dir.')
  elif not args.midi_dir and not args.midi_path:
    raise ValueError('None of midi_dir or midi_path is provided. '
                     'Please provide at least one of midi_dir or midi_path.')
  elif args.midi_dir:
    midi_file_list = glob.glob(args.midi_dir + '/*.mid')
    if len(midi_file_list) == 0:
      raise FileNotFoundError('No midi files found in the directory.')
    for midi_file in tqdm(midi_file_list, desc='Generating files: '):
      synthesize_midi(
        synthesis_generator,
        expression_generator,
        midi_file,
        pitch_offset=args.pitch_offset,
        speed_rate=args.speed_rate,
        output_dir=output_dir,
        sf2_path=args.sf2_path,
        use_fluidsynth=args.use_fluidsynth,
        display_progressbar=False,
        skip_existing_files=args.skip_existing_files,
        save_metadata=args.save_metadata,
      )
  elif args.midi_path:
    synthesize_midi(
      synthesis_generator,
      expression_generator,
      args.midi_path,
      pitch_offset=args.pitch_offset,
      speed_rate=args.speed_rate,
      output_dir=output_dir,
      sf2_path=args.sf2_path,
      use_fluidsynth=args.use_fluidsynth,
      display_progressbar=True,
      skip_existing_files=args.skip_existing_files,
      save_metadata=args.save_metadata,
    )


if __name__ == '__main__':
  main()



"""Utility functions for file io and file path reading."""

#  Copyright 2021 The DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  Lint as: python3

import os
import shutil
import pickle
import json


def get_folder_name(path, num=1):
  """
  Get the name of the folder n levels above the given path.
  Example: a/b/c/d.txt, num=1 -> c, num=2 -> b, ...
  Args:
    path: a file path.
    num: the number of upper directories.

  Returns: the folder name for that level.

  """
  for _ in range(num):
    path = os.path.dirname(path)
  return os.path.basename(path)


def copy_file_to_folder(file_path, dst_dir):
  save_path = os.path.join(dst_dir, os.path.basename(file_path))
  shutil.copy(file_path, save_path)


def pickle_dump(obj, path):
  with open(path, 'wb') as f:
    pickle.dump(obj, f)
    f.close()


def pickle_load(path):
  with open(path, 'rb') as f:
    data = pickle.load(f)
    f.close()
  return data


def json_dump(data_json, json_save_path):
  with open(json_save_path, 'w') as f:
    json.dump(data_json, f)
    f.close()


def json_load(json_path):
  with open(json_path, 'r') as f:
    data = json.load(f)
    f.close()
  return data


def write_str_lines(save_path, lines):
  lines = [l + '\n' for l in lines]
  with open(save_path, 'w', encoding='utf-8') as f:
    f.writelines(lines)



"""Utility functions for audio input, output and spectral operations."""

#  Copyright 2021 The DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  Lint as: python3

import librosa
import librosa.display
import librosa.filters
import numpy as np
import soundfile
import crepe
import tensorflow as tf
import ddsp
from hmmlearn import hmm


def save_wav(wav, path, sample_rate=16000):
  soundfile.write(path, wav, sample_rate)


def load_audio(file_path, sample_rate, mono=True, window_size=0,
               from_numpy=False):
  """
  Load audio file from disk.
  :param file_path: Audio file path.
  :param sample_rate: Sample rate to load, will resample to the sample_rate
  if file has a different sample rate.
  :param mono: Whether to load a mono audio file.
  :param window_size: If larger than 0, then the waveform will be cut off to
  have a length that is multiple of window_size,
  :param from_numpy: If the file_path provided is pointing to a npy file.
  :return:
  """
  if from_numpy:
    y = np.load(file_path)
  else:
    y, _ = librosa.load(file_path, sample_rate, mono=mono, dtype=np.float64)
  if window_size > 0:
    output_length = len(y) // window_size * window_size
    y = y[:output_length]
  return y


def extract_f0(wav, frame_shift_ms=5, sr=44100, unvoice=True, no_log=False):
  """Extract f0 from audio using CREPE."""
  if sr != 16000:
    raise RuntimeError('CREPE method should use sr=16khz')
  _, frequency, confidence, _ = crepe.predict(
    wav, sr=sr,
    viterbi=True,
    step_size=frame_shift_ms,
    verbose=0 if no_log else 1)
  f0 = frequency
  if unvoice:
    is_voiced = predict_voicing(confidence)
    frequency_unvoiced = frequency * is_voiced
    f0 = frequency_unvoiced

  return f0


def predict_voicing(confidence):
  # https://github.com/marl/crepe/pull/26
  """
  Find the Viterbi path for voiced versus unvoiced frames.
  Parameters
  ----------
  confidence : np.ndarray [shape=(N,)]
      voicing confidence array, i.e. the confidence in the presence of
      a pitch
  Returns
  -------
  voicing_states : np.ndarray [shape=(N,)]
      HMM predictions for each frames state, 0 if unvoiced, 1 if
      voiced
  """
  # uniform prior on the voicing confidence
  starting = np.array([0.5, 0.5])

  # transition probabilities inducing continuous voicing state
  transition = np.array([[0.99, 0.01], [0.01, 0.99]])

  # mean and variance for unvoiced and voiced states
  means = np.array([[0.0], [1.0]])
  variances = np.array([[0.25], [0.25]])

  # fix the model parameters because we are not optimizing the model
  model = hmm.GaussianHMM(n_components=2)
  model.startprob_, model.covars_, model.transmat_, model.means_, \
  model.n_features = starting, variances, transition, means, 1

  # find the Viterbi path
  voicing_states = model.predict(confidence.reshape(-1, 1), [len(confidence)])

  return np.array(voicing_states)


def spectral_centroid(wav, hop_length, sr):
  centroid = librosa.feature.spectral_centroid(y=wav, sr=sr,
                                               hop_length=hop_length)
  return centroid


def tf_stft(audio, win_length, hop_length, n_fft, pad_end=True):
  s = tf.signal.stft(
    signals=audio,
    frame_length=win_length,
    frame_step=hop_length,
    fft_length=n_fft,
    pad_end=pad_end)
  mag = tf.abs(s)
  return tf.cast(mag, tf.float32)


def tf_mel(audio, sample_rate, win_length, hop_length, n_fft, num_mels, fmin=40,
           pad_end=True):
  """Calculate Mel Spectrogram."""
  mag = tf_stft(audio, win_length, hop_length, n_fft, pad_end=pad_end)
  num_spectrogram_bins = int(mag.shape[-1])
  hi_hz = sample_rate // 2
  linear_to_mel_matrix = tf.signal.linear_to_mel_weight_matrix(
    num_mels,
    num_spectrogram_bins,
    sample_rate,
    fmin,
    hi_hz)
  mel = tf.tensordot(mag, linear_to_mel_matrix, 1)
  mel.set_shape(mag.shape[:-1].concatenate(linear_to_mel_matrix.shape[-1:]))
  return mel


def tf_log_mel(audio, sample_rate, win_length, hop_length, n_fft, num_mels,
               fmin=40, pad_end=True):
  mel = tf_mel(audio, sample_rate, win_length, hop_length, n_fft, num_mels,
               fmin=fmin, pad_end=pad_end)
  return ddsp.core.safe_log(mel)



#  Copyright 2022 The MIDI-DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Model class for Synthesis Generator + DDSP Inference module."""

import tensorflow as tf
import ddsp
import ddsp.training

from midi_ddsp.utils.inference_utils import get_process_group
from .interpretable_conditioning import extract_harm_controls


class SynthCoder(tf.keras.Model):
  """The DDSP Inference module."""

  def __init__(self, encoder, decoder):
    super().__init__()
    self.encoder = encoder
    self.decoder = decoder
    self.model_type = 'SynthCoder'

  def encode(self, inputs, training=False):
    z = self.encoder(inputs, training=training)
    return z

  def decode(self, z, inputs):
    synth_params = self.decoder([z, inputs])
    return synth_params

  def call(self, inputs, training=None):
    z = self.encode(inputs, training=training)
    synth_params = self.decode(z, inputs)
    return synth_params

  def _build(self, inputs):
    inputs, kwargs = inputs
    self(inputs, **kwargs)


class MIDIExpressionAE(tf.keras.Model):
  """
  The DDSP Inference + Synthesis Generator module.
  The naming of the module is not exactly the same as in the MIDI-DDSP paper,
  but one can find correspondence in following crude diagram:

                                       Note Sequence
                                             │
          Note Expression Controls           │       │
            (conditioning_dict)    ──────────┤       │
                    ▲                        │       │
      Feature       │                        ├────┐  │
     Extraction     │                        │    │  │
  (gen_cond_dict)   │                        ▼    │  │ Synthesis Generator
           Synthesis Parameters             f0    │  │   (midi_decoder)
            (f0,amps,hd,noise)               │    │  │
                    ▲                        ├────┘  │
      DDSP          │                        │       │
   Inference        │                        │       │
  (SynthCoder)      │                        │       │
                    │                        ▼
              Audio Feature         Synthesis Parameters
               (f0,ld,mel)           (f0,amps,hd,noise)
                    ▲                        │
      CREPE         │                        │
   A-Weighting      │                        │ DDSP
     Mel-STFT       │                        │
                    │                        ▼
                  Audio                    Audio
                  Input               Reconstruction
  """

  def __init__(self, synth_coder, midi_decoder, n_frames=1000, frame_size=64,
               sample_rate=16000, reverb_module=None,
               use_f0_ld=False):
    super().__init__()
    self.synth_coder = synth_coder
    self.midi_decoder = midi_decoder
    self.reverb_module = reverb_module
    self.model_type = 'MIDIExpressionAE'

    self.processor_group = get_process_group(n_frames, frame_size, sample_rate,
                                             use_angular_cumsum=False)
    self.n_frames = n_frames
    self.frame_size = frame_size
    self.sample_rate = sample_rate

    self.use_f0_ld = use_f0_ld

  def train_synth_coder_only(self):
    self.midi_decoder.trainable = False
    self.synth_coder.trainable = True

  def freeze_synth_coder(self):
    self.midi_decoder.trainable = True
    self.synth_coder.trainable = False
    if self.reverb_module is not None:
      self.reverb_module.trainable = False

  def run_synth_coder(self, features, training=True):
    synth_params = self.synth_coder(features, training=training)
    control_params = self.processor_group.get_controls(synth_params,
                                                       verbose=False)
    synth_audio = self.processor_group.get_signal(control_params)
    if self.reverb_module is not None:
      synth_audio = self.reverb_module(synth_audio,
                                       reverb_number=features['instrument_id'],
                                       training=training)
    return synth_params, control_params, synth_audio

  @staticmethod
  def get_gt_midi(features):
    """Prepare the ground truth MIDI."""
    f0_midi = ddsp.core.hz_to_midi(features['f0_hz'])
    q_pitch = tf.cast(features['midi'][..., tf.newaxis], tf.float32)

    q_vel = q_pitch * 0.0  # (Jesse) Don't use velocities for now.

    f0_loss_weights = tf.cast(tf.abs(f0_midi - q_pitch) < 2.0, tf.float32)

    return q_pitch, q_vel, f0_loss_weights, features['onsets'], features[
      'offsets']

  def gen_cond_dict_from_feature(self, features, training=False):
    synth_params, control_params, synth_audio = self.run_synth_coder(features,
                                                                     training=
                                                                     training)

    # synth_params_normalized: scaled and normalized synth params
    synth_params_normalized = extract_harm_controls(control_params)

    midi_features = self.get_gt_midi(features)

    conditioning_dict = self.midi_decoder.gen_cond_dict(synth_params_normalized,
                                                        midi_features)

    return synth_params_normalized, midi_features, conditioning_dict

  def gen_audio_from_cond_dict(self, conditioning_dict, midi_features,
                               instrument_id=None,
                               use_angular_cumsum=True,
                               display_progressbar=False):
    z_midi_decoder, params_pred = self.midi_decoder.gen_params_from_cond(
      conditioning_dict, midi_features,
      instrument_id=instrument_id,
      display_progressbar=display_progressbar)

    self.processor_group = get_process_group(z_midi_decoder.shape[1],
                                             self.frame_size, self.sample_rate,
                                             use_angular_cumsum=
                                             use_angular_cumsum)

    if self.use_f0_ld:
      midi_synth_params, midi_control_params, midi_audio = self.run_synth_coder(
        {'f0_hz': params_pred['f0_hz'],
         'loudness_db': params_pred['ld'],
         'instrument_id': instrument_id},
        training=False)
      f0_pred, amps_pred, hd_pred, noise_pred = extract_harm_controls(
        midi_control_params)

    else:
      midi_synth_params = self.processor_group.get_controls(
        {'amplitudes': params_pred['amplitudes'],
         'harmonic_distribution': params_pred[
           'harmonic_distribution'],
         'noise_magnitudes': params_pred['noise_magnitudes'],
         'f0_hz': params_pred['f0_hz'], },
        verbose=False)

      midi_audio = self.processor_group.get_signal(midi_synth_params)
      if self.reverb_module is not None:
        midi_audio = self.reverb_module(midi_audio, reverb_number=instrument_id,
                                        training=False)
      f0_pred, amps_pred, hd_pred, noise_pred = extract_harm_controls(
        midi_synth_params)

    midi_control_params = (f0_pred, amps_pred, hd_pred, noise_pred)
    return midi_audio, midi_control_params, midi_synth_params

  def call(self, features, training=False, run_synth_coder_only=False):
    """Run the network to get a prediction and optionally get losses."""

    synth_params, control_params, synth_audio = self.run_synth_coder(features,
                                                                     training=
                                                                     training)
    if run_synth_coder_only:  # only run synth coder branch
      outputs = {
        'synth_params': synth_params,
        'synth_audio': synth_audio,
      }
      return outputs

    # synth_params_normalized: scaled and normalized synth params
    synth_params_normalized = extract_harm_controls(control_params,
                                                    stop_gradient=True)

    midi_features = self.get_gt_midi(features)

    # --- MIDI Decoding
    conditioning_dict, params_pred = self.midi_decoder(features,
                                                       synth_params_normalized,
                                                       midi_features,
                                                       training=training,
                                                       synth_params=
                                                       control_params)

    if self.use_f0_ld:
      midi_synth_params, midi_control_params, midi_audio = self.run_synth_coder(
        {'f0_hz': params_pred['f0_hz'],
         'loudness_db': params_pred['ld'],
         'instrument_id': features[
           'instrument_id']
         },
        training=training)

      f0_pred, amps_pred, hd_pred, noise_pred = extract_harm_controls(
        midi_control_params)

    else:
      midi_synth_params = self.processor_group.get_controls(
        {'amplitudes': params_pred['amplitudes'],
         'harmonic_distribution': params_pred[
           'harmonic_distribution'],
         'noise_magnitudes': params_pred['noise_magnitudes'],
         'f0_hz': params_pred['f0_hz'], },
        verbose=False)
      # --- MIDI Audio Losses
      midi_audio = self.processor_group.get_signal(midi_synth_params)

      if self.reverb_module is not None:
        midi_audio = self.reverb_module(midi_audio,
                                        reverb_number=features['instrument_id'],
                                        training=training)

      f0_pred, amps_pred, hd_pred, noise_pred = extract_harm_controls(
        midi_synth_params)

    # unpack things
    f0, amps, hd, noise = synth_params_normalized
    q_pitch, q_vel, f0_loss_weights, onsets, offsets = midi_features

    # --- Finalize and return
    outputs = {
      'synth_params': synth_params,
      'control_params': control_params,
      'synth_audio': synth_audio,
      'midi_synth_params': midi_synth_params,
      'midi_audio': midi_audio,
      'q_pitch': q_pitch,
      'q_vel': q_vel,
      'conditioning_dict': conditioning_dict,
      'amps': amps,
      'hd': hd,
      'noise': noise,
      'amps_pred': amps_pred,
      'hd_pred': hd_pred,
      'noise_pred': noise_pred,
      'f0_loss_weights': f0_loss_weights,
      'params_pred': params_pred,
    }

    return outputs

  def _build(self, inputs):
    inputs, kwargs = inputs
    self(inputs, **kwargs)



#  Copyright 2022 The MIDI-DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Create dataset and dataloader for the URMP dataset to train
the Synthesis Generator."""

from ddsp.training.data import UrmpMidi
import math


def get_tfrecord_length(dataloader):
  """Return the length of a dataloader."""
  train_dataset_single = dataloader.get_batch(batch_size=1,
                                              shuffle=False,
                                              repeats=1)
  c = 0
  for _ in train_dataset_single:
    c += 1
  return c


def get_dataset(hp):
  """Return dataloader based on hyperparameters."""
  data_dir = hp.data_dir
  train_data_loader = UrmpMidi(data_dir, instrument_key=hp.instrument,
                               split='train', suffix='batched')
  training_data = train_data_loader.get_batch(batch_size=hp.batch_size,
                                              shuffle=True,
                                              repeats=-1,
                                              drop_remainder=True)
  test_data_loader = UrmpMidi(data_dir, instrument_key=hp.instrument,
                              split='test', suffix='batched')
  evaluation_data = test_data_loader.get_batch(batch_size=hp.batch_size * 3,
                                               shuffle=True,
                                               repeats=1,
                                               drop_remainder=False)

  length_training_data = get_tfrecord_length(train_data_loader)
  length_evaluation_data = get_tfrecord_length(test_data_loader)
  length_training_data = math.ceil(length_training_data / hp.batch_size)
  length_evaluation_data = math.ceil(
    length_evaluation_data / (hp.batch_size * 3))

  training_data = iter(training_data)

  return training_data, length_training_data, evaluation_data, \
         length_evaluation_data



#  Copyright 2022 The MIDI-DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Get Synthesis Generator module from hyperparameters."""

import ddsp.training
from ddsp.training import nn
import tensorflow as tf
from midi_ddsp.data_handling.instrument_name_utils import NUM_INST
from midi_ddsp.modules.model import SynthCoder, MIDIExpressionAE
from midi_ddsp.modules.ddsp_inference import MelF0LDEncoder, F0LDEncoder, \
  FCHarmonicDecoder, FCStackHarmonicDecoder, Cnn8
from midi_ddsp.modules.reverb_modules import ReverbModules
from midi_ddsp.modules.synth_params_decoder import MidiToSynthAutoregDecoder, \
  MidiToF0LDAutoregDecoder, F0LDAutoregDecoder, \
  MidiNoiseToHarmonicDecoder
from midi_ddsp.modules.midi_decoder import ExpressionMidiDecoder, \
  MidiDecoder



def get_midi_decoder(hp):
  """Get midi_decoder from hyperparameters."""
  z_preconditioning_stack = nn.FcStackOut(ch=512, layers=5, n_out=256)
  if hp.midi_decoder_decoder_net == 'dilated_conv':
    net = nn.DilatedConvStack(
      ch=256,
      layers_per_stack=5,
      stacks=4,
      norm_type='layer',
      conditional=True,
    )
    midi_decoder_decoder = ddsp.training.decoders.MidiToHarmonicDecoder(
      net=net,
      f0_residual=True,
      norm=True,
      output_splits=(('f0_midi', 1),
                     ('amplitudes', 1),
                     ('harmonic_distribution', 60),
                     ('noise_magnitudes', 65)),
    )
  elif hp.midi_decoder_decoder_net == 'noise_dilated_conv':
    net = nn.DilatedConvStack(
      ch=128,
      layers_per_stack=5,
      stacks=4,
      norm_type='layer',
      conditional=True,
    )
    midi_decoder_decoder = MidiNoiseToHarmonicDecoder(
      net=net,
      f0_residual=True,
      norm=True,
      output_splits=(('f0_midi', 1),
                     ('amplitudes', 1),
                     ('harmonic_distribution', 60),
                     ('noise_magnitudes', 65)),
    )
  elif hp.midi_decoder_decoder_net == 'rnn_synth_params':
    midi_decoder_decoder = MidiToSynthAutoregDecoder()

  elif hp.midi_decoder_decoder_net == 'rnn_f0_ld':
    midi_decoder_decoder = MidiToF0LDAutoregDecoder()

  if hp.midi_decoder_type == 'interpretable_conditioning':
    midi_decoder = ExpressionMidiDecoder(
      decoder=midi_decoder_decoder,
      decoder_type=hp.midi_decoder_decoder_net,
      z_preconditioning_stack=z_preconditioning_stack,
      multi_instrument=hp.multi_instrument,
      position_code=hp.position_code,
      without_note_expression=hp.without_note_expression,
    )

  elif hp.midi_decoder_type == 'midi_decoder':
    midi_decoder_decoder = MidiToF0LDAutoregDecoder(
      sampling_method='random')
    midi_decoder = MidiDecoder(
      decoder=midi_decoder_decoder,
      multi_instrument=hp.multi_instrument,
    )
  elif hp.midi_decoder_type == 'unconditioned':
    midi_decoder_decoder = F0LDAutoregDecoder(sampling_method='random')
    midi_decoder = MidiDecoder(
      decoder=midi_decoder_decoder,
      multi_instrument=hp.multi_instrument,
    )

  return midi_decoder


def get_synthesis_generator(hp):
  if hp.midi_decoder_decoder_net == 'rnn_f0_ld' or \
        hp.midi_decoder_type == 'midi_decoder' or \
        hp.midi_decoder_type == 'unconditioned':
    use_f0_ld = True
    encoder = F0LDEncoder()
    decoder = FCStackHarmonicDecoder(hp.nhramonic, hp.nnoise)
  else:
    cnn = Cnn8(pool_size=(1, 2))
    encoder = MelF0LDEncoder(cnn, hp.nhid, hp.sample_rate, hp.win_length,
                             hp.hop_length, hp.n_fft, hp.num_mels,
                             hp.fmin)
    decoder = FCHarmonicDecoder(hp.nhramonic, hp.nnoise)
    use_f0_ld = False
  synth_coder = SynthCoder(encoder, decoder)

  midi_decoder = get_midi_decoder(hp)

  if hp.reverb:
    num_reverb = NUM_INST if hp.multi_instrument else 1
    reverb_module = ReverbModules(num_reverb=num_reverb,
                                  reverb_length=hp.reverb_length)
  else:
    reverb_module = None

  model = MIDIExpressionAE(
    synth_coder=synth_coder,
    midi_decoder=midi_decoder,
    n_frames=hp.sequence_length,
    frame_size=hp.frame_size,
    sample_rate=hp.sample_rate,
    reverb_module=reverb_module,
    use_f0_ld=use_f0_ld
  )

  return model


def get_fake_data_synthesis_generator(hp):
  """Get the fake data for building the synthesis generator."""
  fake_data = {
    'audio': tf.random.normal([1, hp.sequence_length * hp.frame_size]),
    'mel': tf.random.normal([1, hp.sequence_length, hp.num_mels]),
    'f0_hz': tf.random.normal([1, hp.sequence_length, 1]),
    'loudness_db': tf.random.normal([1, hp.sequence_length, 1]),
    'midi': tf.ones([1, hp.sequence_length], dtype=tf.int64),
    'instrument_id': tf.ones([1], dtype=tf.int64),
    'onsets': tf.zeros([1, hp.sequence_length], dtype=tf.int64),
    'offsets': tf.zeros([1, hp.sequence_length], dtype=tf.int64),
  }
  return fake_data, {'training': True}



"""Write audio and spectrogram to tensorboard summary file."""
#  Copyright 2021 The DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  Lint as: python3

from ddsp.training.summaries import audio_summary, spectrogram_summary


def write_tensorboard_audio(writer, data, outputs, step, tag):
  """
  Write audio and spectrogram to tensorboard summary file.
  Args:
    writer: writer of the tensorboard summary file.
    data: the data batch from the dataset.
    outputs: outputs from the model output.
    step: current training step.
    tag: tag for the tensorboard summary. Usually is "train" or "test"

  """
  with writer.as_default():
    spectrogram_summary(data['audio'], outputs['synth_audio'], step, name='',
                        tag=f'{tag}/spectrogram_synth_coder')
    audio_summary(data['audio'], step, sample_rate=16000,
                  name=f'{tag}_audio_original')
    audio_summary(outputs['synth_audio'], step, sample_rate=16000,
                  name=f'{tag}_audio_synth_coder')
    if 'midi_audio' in outputs.keys():
      spectrogram_summary(data['audio'], outputs['midi_audio'], step, name='',
                          tag=f'{tag}/spectrogram_midi_decoder')
      audio_summary(outputs['midi_audio'], step, sample_rate=16000,
                    name=f'{tag}_audio_midi_decoder')



#  Copyright 2022 The MIDI-DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The discriminator for Synthesis Generator."""

import tensorflow as tf

tfk = tf.keras
tfkl = tfk.layers

conv_init = tf.keras.initializers.RandomNormal(
  mean=0.0, stddev=0.02
)


class DBlock(tfkl.Layer):
  """Single discriminator block."""

  def __init__(self, nhid):
    super().__init__()
    self.cond_input_conv = tfkl.Conv1D(nhid, 3, 2, 'same',
                                       kernel_initializer=conv_init)
    self.spec_input_conv = tfkl.Conv1D(nhid, 3, 2, 'same',
                                       kernel_initializer=conv_init)

    self.cond_output_conv = tfkl.Conv1D(nhid, 3, 1, 'same',
                                        kernel_initializer=conv_init)
    self.spec_output_conv = tfkl.Conv1D(nhid, 3, 1, 'same',
                                        kernel_initializer=conv_init)

    self.leaky_relu = tfkl.LeakyReLU(0.2)

    self.cond_layer_norm = tfkl.LayerNormalization()
    self.spec_layer_norm = tfkl.LayerNormalization()

  def call(self, inputs):
    cond_features, spectral_features = inputs

    cond_z_in = self.cond_input_conv(cond_features)
    spec_z_in = self.spec_input_conv(spectral_features)
    spec_z_in += cond_z_in

    cond_z_out = self.leaky_relu(cond_z_in)
    spec_z_out = self.leaky_relu(spec_z_in)

    cond_z_out = self.cond_layer_norm(
      self.cond_output_conv(cond_z_out) + cond_z_in)
    spec_z_out = self.spec_layer_norm(
      self.spec_output_conv(spec_z_out) + spec_z_in)

    return cond_z_out, spec_z_out


class NLayerDiscriminator(tfkl.Layer):
  """A discriminator in a single level, containing multiple blocks."""

  def __init__(self, num_dblock=4, nhid=128):
    super().__init__()
    self.num_dblock = num_dblock
    self.dblock = {str(i): DBlock(nhid) for i in range(num_dblock)}
    self.out_proj = tfkl.Conv1D(1, 3, 1, 'same', kernel_initializer=conv_init)

  def call(self, inputs):
    feature_maps = []
    outputs = inputs
    for i in range(self.num_dblock):
      outputs = self.dblock[str(i)](outputs)
      feature_maps.append(outputs[1])

    out_score = self.out_proj(tf.concat(outputs, -1))
    feature_maps.append(out_score)

    return feature_maps


class Discriminator(tfkl.Layer):
  """The multi-scale discriminator, has num_D discriminators at num_D scale,
   each discriminator consists of num_dblock blocks."""

  def __init__(self, num_D=3, num_dblock=4, nhid=128):
    super().__init__()
    self.model = {}
    for i in range(num_D):
      self.model[f'disc_{i}'] = NLayerDiscriminator(
        num_dblock, nhid
      )

    self.downsample = tfkl.AveragePooling1D(4, 2, padding='same')

  def call(self, inputs):
    results = []
    for _, disc in self.model.items():
      results.append(disc(inputs))
      inputs = (self.downsample(inputs[0]), self.downsample(inputs[1]))
    return results



#  Copyright 2022 The MIDI-DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Calculate note expression controls from synthesis parameters."""

import tensorflow as tf
import ddsp
import ddsp.training
import math


def extract_harm_controls(synth_params, log_scale=True, stop_gradient=False):
  """Get harmonic synth controls from the outputs of the processor group."""
  f0 = synth_params['harmonic']['controls']['f0_hz']
  amps = synth_params['harmonic']['controls']['amplitudes']
  hd = synth_params['harmonic']['controls']['harmonic_distribution']
  noise = synth_params['filtered_noise']['controls']['magnitudes']

  if log_scale:
    amps = ddsp.core.amplitude_to_db(amps, use_tf=True)
    noise = ddsp.core.amplitude_to_db(noise, use_tf=True)

  if stop_gradient:
    amps = tf.stop_gradient(amps)
    hd = tf.stop_gradient(hd)
    noise = tf.stop_gradient(noise)

  return f0, amps, hd, noise


def get_pitch_deviation(f0_midi, f0, mask_large_diff=True):
  # f0_midi: [batch_size,T,1]
  f0_midi_midi_scale = ddsp.core.hz_to_midi(f0_midi)
  f0_midi_scale = ddsp.core.hz_to_midi(f0)
  pitch_deviation = f0_midi_midi_scale - f0_midi_scale
  if mask_large_diff:
    pitch_deviation = tf.where(tf.greater(tf.abs(pitch_deviation), 2.0), 0.0,
                               pitch_deviation)
  return pitch_deviation


def get_amplitude(amplitude):
  # amplitude: [batch_size,T,1]
  return tf.convert_to_tensor(amplitude)


def get_brightness(harmonic_distribution):
  # weighted average across harmonic_distribution
  # harmonic_distribution: [batch, nframe, nharmonics],
  # already scaled and normalized
  nharmonics = harmonic_distribution.shape[-1]
  num_bin = tf.cast(tf.linspace(1, nharmonics, num=nharmonics), tf.float32)[
            tf.newaxis, tf.newaxis, :]
  return tf.reduce_mean(harmonic_distribution * num_bin, axis=-1, keepdims=True)


def get_noise_level(noise_magnitudes):
  # noise_magnitudes: [n_batch, n_frames, n_frequencies]
  noise_amp = tf.reduce_mean(noise_magnitudes, axis=-1, keepdims=True)
  return noise_amp


def get_normal_window(t, alpha=0.5):
  w = 0.5 * (1 - tf.math.cos(2 * math.pi * t / alpha))
  mask = tf.cast(tf.logical_and(t > 0.5 * alpha, t < (1 - 0.5 * alpha)),
                 tf.float32)
  w = tf.clip_by_value(w + mask, 0, 1)
  return w


def get_vibrato_feature(pitch_deviation,
                        note_mask,
                        sampling_interval=0.004,
                        min_note_length=50,
                        vibrato_rate_min=3,
                        vibrato_rate_max=9):
  batch_size = pitch_deviation.shape[0]
  total_length = pitch_deviation.shape[1]
  pitch_deviation_masked = note_mask * pitch_deviation
  pv_mean = ddsp.training.nn.pool_over_notes(pitch_deviation, note_mask,
                                             return_std=False)
  pitch_deviation_masked = note_mask * (
        pitch_deviation_masked - pv_mean)  # filter out DC component

  each_note_idx = tf.cumsum(note_mask, axis=1) * tf.cast(~(note_mask == 0),
                                                         tf.float32)
  each_note_len = tf.reduce_sum(note_mask, axis=1, keepdims=True)
  each_note_time_ratio = each_note_idx / each_note_len
  window = get_normal_window(each_note_time_ratio)
  pitch_deviation_masked *= window

  # [batch_size*max_regions, n_frames]
  pitch_deviation_masked = tf.reshape(
    tf.transpose(pitch_deviation_masked, [0, 2, 1]), [-1, total_length])

  T = sampling_interval
  N = pitch_deviation.shape[1]
  f = tf.linspace(0, int(1 / T), N)

  s_vibrato = tf.abs(tf.signal.rfft(pitch_deviation_masked))
  s_vibrato = tf.math.divide_no_nan(s_vibrato,
                                    tf.reshape(each_note_len, [-1, 1]))

  vibrato_rate_idx = tf.argmax(tf.cast(s_vibrato, tf.float32), -1)

  vibrato_rate = tf.cast(tf.gather(f, vibrato_rate_idx), tf.float32)
  vibrato_extend = tf.gather_nd(s_vibrato, vibrato_rate_idx[:, tf.newaxis],
                                batch_dims=1)
  # replace nan caused by rfft zeros with 0
  vibrato_extend = tf.where(tf.math.is_nan(vibrato_extend), 0, vibrato_extend)

  # filter out vibrato between 3-9 hz
  vibrato_mask = tf.math.logical_and(vibrato_rate >= vibrato_rate_min,
                                     vibrato_rate <= vibrato_rate_max)
  # filter out vibrato extend > 0.012
  # vibrato_mask = tf.math.logical_and(vibrato_mask, vibrato_extend > 0.012)

  # note length > 50 frames
  vibrato_mask = tf.math.logical_and(vibrato_mask,
                                     tf.reshape(each_note_len, [-1]) >
                                     min_note_length)

  # vibrato more than one cycle
  more_than_one_cycle_mask = vibrato_rate > tf.math.divide_no_nan(
    1., tf.reshape(each_note_len, [-1]) * sampling_interval)
  vibrato_mask = tf.math.logical_and(vibrato_mask, more_than_one_cycle_mask)
  vibrato_mask = tf.cast(vibrato_mask, tf.float32)

  # construct output
  vibrato_extend = vibrato_mask * vibrato_extend
  vibrato_rate = vibrato_mask * vibrato_rate
  frame_wise_vibrato_rate = tf.reduce_sum(
    tf.reshape(vibrato_rate, [batch_size, 1, -1]) * note_mask, axis=-1,
    keepdims=True)
  frame_wise_vibrato_extend = tf.reduce_sum(
    tf.reshape(vibrato_extend, [batch_size, 1, -1]) * note_mask, axis=-1,
    keepdims=True)

  return frame_wise_vibrato_rate, frame_wise_vibrato_extend


def get_amplitudes_max_pos(amplitudes, note_mask):
  note_mask_reverse = tf.cast(tf.logical_not(tf.cast(note_mask, tf.bool)),
                              tf.float32)
  amplitudes_masked = note_mask * amplitudes + note_mask_reverse * -1000
  # multiply -1000 here preventing argmax to the mask
  each_note_idx = tf.cumsum(note_mask, axis=1) * tf.cast(~(note_mask == 0),
                                                         tf.float32)
  each_note_len = tf.reduce_max(each_note_idx, axis=1, keepdims=True)
  note_onset_index = tf.argmax(note_mask, axis=1)

  # index inside a note that achieves max amplitudes
  amplitudes_max_idx = tf.argmax(amplitudes_masked, axis=1) - note_onset_index
  amplitudes_max_pos = tf.math.divide_no_nan(
    tf.cast(amplitudes_max_idx[:, tf.newaxis, :], tf.float32), each_note_len)
  amplitudes_max_pos = tf.reduce_sum(amplitudes_max_pos * note_mask, axis=-1,
                                     keepdims=True)
  return amplitudes_max_pos


def get_attack_level(noise_level, note_mask):
  each_note_idx = tf.cumsum(note_mask, axis=1) * tf.cast(~(note_mask == 0),
                                                         tf.float32)
  attack_mask = tf.cast(tf.logical_and(each_note_idx > 0, each_note_idx <= 10),
                        tf.float32)  # pool over first 10 frames
  # [b, n, d]
  attack_notes_mean = ddsp.training.nn.get_note_moments(noise_level,
                                                        attack_mask,
                                                        return_std=False)
  # [b, t, n, d]
  attack_time_notes_mean = (attack_notes_mean[:, tf.newaxis, ...] *
                            note_mask[..., tf.newaxis])
  # [b, t, d]
  attack_level = tf.reduce_sum(attack_time_notes_mean, axis=2)
  return attack_level


def get_interpretable_conditioning(f0_midi, f0, amplitude,
                                   harmonic_distribution, noise_magnitudes):
  """Calculate conditioning from synthesis needed for
  calculating note expression controls."""
  pitch_deviation = get_pitch_deviation(f0_midi, f0)
  amplitude = get_amplitude(amplitude)
  brightness = get_brightness(harmonic_distribution)
  noise_level = get_noise_level(noise_magnitudes)
  conditioning = {
    'pitch_deviation': pitch_deviation,
    'amplitude': amplitude,
    'brightness': brightness,
    'noise_level': noise_level,
  }
  return conditioning


def adsr_get_note_mask(q_pitch, max_regions=200, note_on_only=True):
  """Get a binary mask for each note from a monophonic instrument based on
  ADSR model. This function is modified from ddsp.training.nn.get_note_mask.
  In addition to the note boundary, it individually pool the first (A)
  and last few frames (D+R) of a note.

  Each transition of the value creates a new region. Returns the mask of each
  region.
  Args:
    q_pitch: A quantized value, such as pitch or velocity. Shape
      [batch, n_timesteps] or [batch, n_timesteps, 1].
    max_regions: Maximum number of note regions to consider in the sequence.
      Also, the channel dimension of the output mask. Each value transition
      defines a new region, e.g. each note-on and note-off count as a separate
      region.
    note_on_only: Return a mask that is true only for regions where the pitch
      is greater than 0.

  Returns:
    A binary mask of each region [batch, n_timesteps, max_regions].
  """
  # Only batch and time dimensions.
  if len(q_pitch.shape) == 3:
    q_pitch = q_pitch[:, :, 0]

  # Get onset and offset points.
  edges = tf.abs(ddsp.core.diff(q_pitch, axis=1)) > 0

  # Count endpoints as starts/ends of regions.
  edges = edges[:, :-1, ...]
  edges = tf.pad(edges, [[0, 0], [1, 0]], mode='constant', constant_values=True)
  edges = tf.pad(edges, [[0, 0], [0, 1]], mode='constant',
                 constant_values=False)
  edges = tf.cast(edges, tf.int32)

  # Count up onset and offsets for each timestep.
  # Assumes each onset has a corresponding offset.
  # The -1 ensures that the 0th index is the first note.
  edge_idx = tf.cumsum(edges, axis=1) - 1

  # Create masks of shape [batch, n_timesteps, max_regions].
  note_mask = edge_idx[..., None] == tf.range(max_regions)[None, None, :]
  note_mask = tf.cast(note_mask, tf.float32)

  if note_on_only:
    # [batch, notes]
    note_pitches = ddsp.training.nn.get_note_moments(q_pitch, note_mask,
                                                     return_std=False)
    # [batch, time, notes]
    note_on = tf.cast(note_pitches > 0.0, tf.float32)[:, None, :]
    # [batch, time, notes]
    note_mask *= note_on
    note_on_time_dim = tf.reduce_sum(note_mask,
                                     axis=-1)  # note_on in time dimension

  # frame index for each note
  each_note_idx = tf.cumsum(note_mask, axis=1) * tf.cast(~(note_mask == 0),
                                                         tf.float32)
  each_note_idx_reverse = tf.cumsum(note_mask, axis=1, reverse=True) * tf.cast(
    ~(note_mask == 0), tf.float32)
  each_note_len = tf.reduce_max(each_note_idx, axis=1, keepdims=True) * tf.cast(
    each_note_idx > 0, tf.float32)

  each_note_idx_reduce = tf.reduce_sum(each_note_idx, axis=-1)
  each_note_idx_reverse_reduce = tf.reduce_sum(each_note_idx_reverse, axis=-1)
  each_note_len_reduce = tf.reduce_sum(each_note_len, axis=-1)

  attack_mask = tf.math.logical_and(each_note_idx_reduce == 10,
                                    each_note_len_reduce >= 50)
  decay_mask = tf.math.logical_and(each_note_idx_reverse_reduce == 10,
                                   each_note_len_reduce >= 50)

  edges_adsr = edges + tf.cast(attack_mask, tf.int32) + tf.cast(decay_mask,
                                                                tf.int32)
  edge_idx_adsr = tf.cumsum(edges_adsr, axis=1) - 1

  # Create masks of shape [batch, n_timesteps, max_regions].
  note_mask_adsr = edge_idx_adsr[..., None] == tf.range(max_regions)[None, None,
                                               :]
  note_mask_adsr = tf.cast(note_mask_adsr, tf.float32)
  if note_on_only:
    note_mask_adsr *= note_on_time_dim[..., tf.newaxis]

  return note_mask_adsr


def get_conditioning_dict(conditioning, q_pitch, onsets,
                          pool_type='note_pooling'):
  """Calculate note expression controls."""
  # conditioning: dict of conditioning
  if pool_type == 'note_pooling':
    note_mask = ddsp.training.nn.get_note_mask_from_onset(q_pitch, onsets)
  elif pool_type == 'adsr_note_pooling':
    note_mask = adsr_get_note_mask(q_pitch)

  amp_mean, amp_std = ddsp.training.nn.pool_over_notes(
    conditioning['amplitude'], note_mask, return_std=True)
  brightness = ddsp.training.nn.pool_over_notes(conditioning['brightness'],
                                                note_mask, return_std=False)
  attack_level = get_attack_level(conditioning['noise_level'], note_mask)
  vibrato_rate, vibrato_extend = get_vibrato_feature(
    conditioning['pitch_deviation'], note_mask)
  amplitudes_max_pos = get_amplitudes_max_pos(conditioning['amplitude'],
                                              note_mask)

  # scale conditioning so that most value are in [0, 1]
  # TODO: (yusongwu) enable automatic scaling
  amp_mean = tf.where(tf.equal(amp_mean, 0.0), 0.0, amp_mean / 60 + 1.5)
  amp_std *= (2.5 / 60)
  vibrato_extend *= 10
  brightness *= 5
  attack_level = tf.where(tf.equal(attack_level, 0.0), 0.0,
                          attack_level / 40 + 2.625)

  conditioning_dict = {
    'amplitude_mean': amp_mean,
    'amplitude_std': amp_std,
    'vibrato_extend': vibrato_extend,
    'brightness': brightness,
    'attack_level': attack_level,
    # 'vibrato_rate': vibrato_rate,
    'amplitudes_max_pos': amplitudes_max_pos,
  }
  return conditioning_dict



#  Copyright 2022 The MIDI-DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Abstract model for conditional autoregressive RNN used in MIDI-DDSP."""

import tensorflow as tf
from tqdm.autonotebook import tqdm

tfk = tf.keras
tfkl = tfk.layers


class StackedRNN(tfkl.Layer):
  """Stacked RNN implementated using tfkl.Layers."""

  def __init__(self, nhid=256, nlayers=2, rnn_type='gru', dropout=0.5):
    super().__init__()
    if rnn_type == 'gru':
      rnn_layer = tfkl.GRU
    elif rnn_type == 'lstm':
      rnn_layer = tfkl.LSTM
    else:
      raise TypeError('Unknown rnn_type')
    self.nhid = nhid
    self.nlayers = nlayers
    self.net = {
      str(i): rnn_layer(nhid, return_sequences=True, return_state=True,
                        dropout=dropout) for i in
      range(nlayers)}

  def call(self, x, initial_state=None, training=False):
    states_out_all = []
    z_out = x
    if initial_state is None:
      initial_state = [None for _ in range(self.nlayers)]
    for i in range(self.nlayers):
      z_out, states_out = self.net[str(i)](z_out,
                                           initial_state=initial_state[i],
                                           training=training)
      states_out_all.append(states_out)
    return z_out, states_out_all


class TwoLayerCondAutoregRNN(tfkl.Layer):
  """Conditional two-layer autoregressive RNN.
  The RNN here takes input from a conditioning sequence and its previous output.
  The RNN is trained using teacher forcing and inference in autoregressive mode.
  """

  def __init__(self, nhid, n_out, input_dropout=True, input_dropout_p=0.5,
               dropout=0.5, rnn_type='gru'):
    """Constructor."""
    super().__init__()
    self.n_out = n_out
    self.nhid = nhid
    self.input_dropout = input_dropout
    self.input_dropout_p = input_dropout_p
    if rnn_type == 'gru':
      self.rnn1 = tfkl.GRU(nhid, return_sequences=True, return_state=True,
                           dropout=dropout)
      self.rnn2 = tfkl.GRU(nhid, return_sequences=True, return_state=True,
                           dropout=dropout)
    elif rnn_type == 'lstm':
      self.rnn1 = tfkl.LSTM(nhid, return_sequences=True, return_state=True,
                            dropout=dropout)
      self.rnn2 = tfkl.LSTM(nhid, return_sequences=True, return_state=True,
                            dropout=dropout)
    else:
      raise ValueError('Unknown RNN type.')

  def _one_step(self, curr_cond, prev_out, prev_states, training=False):
    """One step inference."""
    prev_states_1, prev_states_2 = prev_states
    curr_z_in = tf.concat([curr_cond, prev_out], -1)  # [batch_size, 1, dim]
    curr_z_in = self.encode_z(curr_z_in, training=training)
    curr_z_out, curr_states_1 = self.rnn1(curr_z_in,
                                          initial_state=prev_states_1,
                                          training=training)
    curr_z_out, curr_states_2 = self.rnn2(curr_z_out,
                                          initial_state=prev_states_2,
                                          training=training)
    curr_out = self.decode_out(curr_z_out)
    return curr_out, (curr_states_1, curr_states_2)

  def autoregressive(self, cond, training=False, display_progressbar=False):
    """Autoregressive inference."""
    cond_encode = self.encode_cond(cond)
    batch_size = cond_encode.shape[0]
    length = cond_encode.shape[1]
    prev_out = tf.tile([[0.0]], [batch_size, self.n_out])[:, tf.newaxis,
               :]  # go_frame
    prev_states = (None, None)
    overall_outputs = []

    for i in tqdm(range(length), position=0, leave=True, desc='Generating: ',
                  disable=not display_progressbar):
      curr_cond = cond_encode[:, i, :][:, tf.newaxis, :]
      prev_out = self.encode_out(prev_out, training=training)
      curr_out, curr_states = self._one_step(curr_cond, prev_out, prev_states,
                                             training=training)
      curr_out = self.sample_out(curr_out, training=training)
      overall_outputs.append(curr_out)
      prev_out, prev_states = curr_out['curr_out'], curr_states

    outputs = {}
    for k in curr_out.keys():
      outputs[k] = tf.concat([x[k] for x in overall_outputs], 1)
    return outputs

  def right_shift_encode_out(self, out, training=False):
    """Right shift the ground truth target by one timestep and encode."""
    go_frame = tf.tile([[0.0]], [out.shape[0], out.shape[-1]])[:, tf.newaxis, :]
    out = tf.concat([go_frame, out[:, :-1, :]], 1)
    out = self.encode_out(out, training=training)
    # input dropout
    if self.input_dropout:
      input_dropout_mask = tf.random.uniform(
        [out.shape[0], out.shape[1], 1]) > self.input_dropout_p
      input_dropout_mask = tf.cast(input_dropout_mask, tf.float32)
      out = out * input_dropout_mask
    return out

  def teacher_force(self, cond, out, training=True):
    """Run teacher force."""
    out_shifted = self.right_shift_encode_out(out, training=training)
    cond = self.encode_cond(cond, training=training)
    z_in = tf.concat([cond, out_shifted], -1)
    z_in = self.encode_z(z_in, training=training)
    z_out, *states = self.rnn1(z_in, training=training)
    z_out, *states = self.rnn2(z_out, training=training)
    output = self.decode_out(z_out, training=training)
    output = self.split_teacher_force_output(output)
    return output

  def encode_z(self, z, training=False):
    return z

  def encode_out(self, out, training=False):
    return out

  def encode_cond(self, cond, training=False):
    return cond

  def decode_out(self, z_out, training=False):
    return z_out

  def sample_out(self, out, training=False):
    return out

  def preprocess(self, cond, out):
    return cond, out

  def postprocess(self, outputs, cond, training=False):
    return outputs

  def split_teacher_force_output(self, output):
    return output

  def call(self, cond, out=None, training=False):
    """Forward call."""
    if training:
      cond, out = self.preprocess(cond, out)
      outputs = self.teacher_force(cond, out, training=training)
    else:
      cond, out = self.preprocess(cond, out)
      outputs = self.autoregressive(cond, training=training)

    outputs = self.postprocess(outputs, cond, training=training)

    return outputs



#  Copyright 2022 The MIDI-DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The hyperparameters for training Synthesis Generator.
For details of each hyperparameters, please see the argpars """

class hparams:
  """Hyperparameters for training Synthesis Generator."""
  # Learning parameters
  batch_size = 4
  clip_grad = 2.5
  lr = 3e-4
  seed = 1111

  # Training parameters
  training_steps = 100000
  log_interval = 100
  checkpoint_save_interval = 5000
  eval_interval = 5000
  mode = 'train'  # train, eval
  data_dir = None
  restore_path = None

  # Model Parameters
  nhid = 256
  sequence_length = 1000
  train_synth_coder_first = True
  midi_audio_loss = True
  add_synth_loss = False
  synth_params_loss = False
  midi_decoder_type = 'interpretable_conditioning'
  position_code = 'index_length'
  midi_decoder_decoder_net = 'rnn_synth_params'
  multi_instrument = True
  instrument = 'vn'
  synth_coder_training_steps = 300
  lambda_recon = 1.0
  use_gan = True
  lambda_G = 1
  sg_z = True
  lr_disc = 1e-4
  write_tfrecord_audio = False
  without_note_expression = False
  discriminator_dim = 256

  # Synthesis & DSP parameters
  nhramonic = 60
  nnoise = 65
  reverb = True
  reverb_length = 48000
  num_mels = 64
  n_fft = 1024
  sample_rate = 16000
  frame_size = 64
  hop_length = frame_size
  win_length = hop_length * 2
  frame_shift_ms = 1000 / sample_rate * frame_size
  fmin = 40



"""Utility functions for creating dataset for expression generator dataset."""

#  Copyright 2021 The DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  Lint as: python3

import numpy as np
import ddsp
import tensorflow as tf
import copy
import glob
from tqdm.autonotebook import tqdm
import os
import matplotlib.pyplot as plt
from ddsp.training.data import UrmpMidiUnsegmented
from midi_ddsp.utils.file_utils import pickle_dump
from midi_ddsp.utils.inference_utils import to_length, make_same_length, get_process_group
from midi_ddsp.utils.audio_io import save_wav
from midi_ddsp.modules.interpretable_conditioning import \
  get_interpretable_conditioning, get_amplitudes_max_pos, get_vibrato_feature, \
  extract_harm_controls


# Sample the whole piece in URMP dataset.

def save_synth_params(data, synth_params, output_dir):
  recording_id = data['recording_id'][0].numpy().decode()
  synth_params = {k: v.numpy()[0] for k, v in synth_params.items()}
  synth_params_save_path = os.path.join(output_dir,
                                        f'{recording_id}.synth_params')
  np.save(synth_params_save_path, synth_params)
  midi_save_path = os.path.join(output_dir, f'{recording_id}.midi')
  np.save(midi_save_path, data['midi'].numpy()[0])
  onsets_save_path = os.path.join(output_dir, f'{recording_id}.onsets')
  np.save(onsets_save_path, data['onsets'].numpy()[0])
  offsets_save_path = os.path.join(output_dir, f'{recording_id}.offsets')
  np.save(offsets_save_path, data['offsets'].numpy()[0])
  instrument_id_save_path = os.path.join(output_dir,
                                         f'{recording_id}.instrument_id')
  np.save(instrument_id_save_path, data['instrument_id'].numpy()[0])
  wav_path = os.path.join(output_dir, f'{recording_id}.wav')
  np.save(wav_path, data['audio'].numpy()[0])
  save_wav(data['synth_audio'].numpy()[0],
           os.path.join(output_dir, f'{recording_id}_synth.wav'), 16000)
  assert len(synth_params['f0']) == len(synth_params['amps'])


def generate_and_save_synth_params(synthesis_generator, dataset, output_dir):
  for data in tqdm(dataset):
    if data['f0_hz'].shape[1] != data['loudness_db'].shape[1]:
      data['loudness_db'] = make_same_length(data['loudness_db'], data['f0_hz'])

    synth_params = synthesis_generator.synth_coder(data, training=False)

    # transfer to normalized synth_params
    processor_group = get_process_group(data['f0_hz'].shape[1])

    synth_params = processor_group.get_controls(synth_params, verbose=False)
    synth_audio = processor_group.get_signal(synth_params)
    synth_audio = synthesis_generator.reverb_module(synth_audio,
                                                    reverb_number=data[
                                                      'instrument_id'])
    data['synth_audio'] = synth_audio

    f0, amps, hd, noise = extract_harm_controls(synth_params)
    synth_params = {'f0': f0, 'amps': amps, 'hd': hd, 'noise': noise}

    save_synth_params(data, synth_params, output_dir)


# Create dataset contains synthesis parameters.

def get_piece_note_features(data, conditioning):
  note_feature_piece = []
  onset_index = np.where(data['onsets'] == 1)[0]
  offset_index = np.where(data['offsets'] == 1)[0]
  assert len(onset_index) == len(offset_index)
  total_frame = conditioning['amplitude'].shape[1]
  last_off = -1

  for on, off in zip(onset_index, offset_index):
    if on != off:
      amp_mean = np.mean(conditioning['amplitude'][0, on:off + 1, :])
      amp_std = np.std(conditioning['amplitude'][0, on:off + 1, :])
      brightness = np.mean(conditioning['brightness'][0, on:off + 1, :])
      attack_level = np.mean(
        conditioning['noise_level'][0, on:min(off + 1, on + 10), :])

      pv = conditioning['pitch_deviation'][0, on:off + 1, :]
      note_mask = tf.ones([1, pv.shape[0], 1])
      note_mask_pad = to_length(note_mask, dst_length=1000)
      pv_pad = to_length(pv[tf.newaxis, ...], dst_length=1000)
      _, frame_wise_vibrato_extend = get_vibrato_feature(pv_pad, note_mask_pad)
      vibrato_extend = frame_wise_vibrato_extend[0, 0, 0].numpy()

      amps = conditioning['amplitude'][0, on:off + 1, :]
      note_mask = tf.ones([1, pv.shape[0], 1])
      note_mask_pad = to_length(note_mask, dst_length=1000)
      amps_pad = to_length(amps[tf.newaxis, ...], dst_length=1000)
      amplitudes_max_pos = get_amplitudes_max_pos(amps_pad, note_mask_pad)
      amplitudes_max_pos = amplitudes_max_pos[0, 0, 0].numpy()

      # scale conditioning so that most value are in [0, 1]
      # TODO: (yusongwu) enable automatic scaling and merge with
      #  get_conditioning_dict
      amp_mean = np.where(np.equal(amp_mean, 0.0), 0.0,
                          amp_mean / 60 + 1.5).item()
      amp_std *= (2.5 / 60)
      vibrato_extend *= 10
      brightness *= 5
      attack_level = np.where(np.equal(attack_level, 0.0), 0.0,
                              attack_level / 40 + 2.625).item()

      note_feature_dict = {
        'amplitude_mean': amp_mean,
        'amplitude_std': amp_std,
        'vibrato_extend': vibrato_extend,
        'brightness': brightness,
        'attack_level': attack_level,
        'amplitudes_max_pos': amplitudes_max_pos,
        'note_length': off - on + 1,
        'note_pitch': data['midi'][on],
        'instrument_id': data['instrument_id'].item()
      }

      assert note_feature_dict['note_pitch'] != 0

      if on - last_off == 0:
        note_feature_dict['note_length'] -= 1

      elif on - last_off > 1:  # add rest note
        note_feature_dict_rest = {k: 0 for k in note_feature_dict}
        note_feature_dict_rest['note_length'] = on - last_off - 1
        note_feature_dict_rest['note_pitch'] = 0
        note_feature_dict_rest['instrument_id'] = note_feature_dict[
          'instrument_id']
        note_feature_piece.append(note_feature_dict_rest)

      last_off = off

      if note_feature_dict['note_length'] != 0:
        note_feature_piece.append(note_feature_dict)

  if last_off < (total_frame - 1):  # add the last rest note
    note_feature_dict_rest = {k: 0 for k in note_feature_dict}
    note_feature_dict_rest['note_length'] = total_frame - last_off - 1
    note_feature_dict_rest['note_pitch'] = 0
    note_feature_dict_rest['instrument_id'] = note_feature_dict['instrument_id']
    note_feature_piece.append(note_feature_dict_rest)

  total_frame_note = sum([note['note_length'] for note in note_feature_piece])
  assert total_frame_note == total_frame

  return note_feature_piece


def note_features_to_input_dict(note_features):
  note_features = copy.deepcopy(note_features)
  note_pitch = note_features['note_pitch']
  note_length = note_features['note_length']
  instrument_id = note_features['instrument_id']
  note_features.pop('note_pitch')
  note_features.pop('note_length')
  note_features.pop('instrument_id')
  conditioning_feature = np.array(list(note_features.values()))
  input_dict = {
    'note_pitch': note_pitch,
    'note_length': note_length,
    'conditioning_feature': conditioning_feature,
    'instrument_id': instrument_id,
  }
  return input_dict


def segment_single(data, window_length, hop_length):
  if len(data) < window_length:
    return [data]
  data_list = [data[i:i + window_length] for i in
               range(0, len(data) - window_length + hop_length, hop_length)]
  if len(data_list[-1]) < window_length:
    # set the last one as same length disregarding hop_length
    data_list[-1] = data[len(data) - window_length:]
  return data_list


def segment_data(note_features_all, window_length=64, hop_length=1):
  data_segmented_all = []
  for note_feature_piece in note_features_all:
    input_dict_piece = [note_features_to_input_dict(n) for n in
                        note_feature_piece]
    data_segmented = segment_single(input_dict_piece,
                                    window_length=window_length,
                                    hop_length=hop_length)
    data_segmented_all.extend(data_segmented)
  return data_segmented_all


def get_all_note_features(data_dir):
  midi_file_list = glob.glob(data_dir + '/*.midi.npy')
  note_features_all = []
  audio_all = []
  for midi_file in tqdm(midi_file_list):
    data = np.load(midi_file.replace('.midi', '.synth_params'),
                   allow_pickle=True).item()
    data.update({'midi': np.load(midi_file)})
    data.update({'onsets': np.load(midi_file.replace('.midi', '.onsets'))})
    data.update({'offsets': np.load(midi_file.replace('.midi', '.offsets'))})
    data.update(
      {'instrument_id': np.load(midi_file.replace('.midi', '.instrument_id'))})
    conditioning = get_interpretable_conditioning(
      ddsp.core.midi_to_hz(data['midi'][np.newaxis, :, np.newaxis],
                           midi_zero_silence=True),
      data['f0'][np.newaxis, ...],
      data['amps'][np.newaxis, ...],
      data['hd'][np.newaxis, ...],
      data['noise'][np.newaxis, ...])
    note_feature_piece = get_piece_note_features(data, conditioning)
    note_features_all.append(note_feature_piece)
    audio_all.append(np.load(midi_file.replace('.midi', '.wav')))
  return note_features_all, audio_all


def get_stat_str(value_list):
  return f'Max: {np.max(value_list):.2f}, ' \
         f'Min: {np.min(value_list):.2f}, ' \
         f'Mean: {np.mean(value_list):.2f}, ' \
         f'Std: {np.std(value_list):.2f}'


def plot_save_expression_stats(key, note_params_reduced, show_plot=True,
                               save_fig=False, output_dir=None):
  value_list = [n[key] for n in note_params_reduced]
  print(f'{key} {get_stat_str(value_list)}')
  plt.hist(value_list, bins=50)
  if show_plot:
    plt.show()
  if save_fig:
    plt.savefig(os.path.join(output_dir, f'{key}_stat.png'))
  # if show_plot or save_fig:
  #   plt.close()


def plot_stats(note_params_reduced, instrument_id=None, show_plot=True,
               save_fig=False, output_dir=None):
  if save_fig and not output_dir:
    raise ValueError(
      'Please specify output directory for saving plot statistics.')
  if instrument_id is None:
    note_params_reduced = [n for n in note_params_reduced if
                           n['note_pitch'] != 0]
  else:
    note_params_reduced = [n for n in note_params_reduced if
                           (n['note_pitch'] != 0 and n[
                             'instrument_id'] == instrument_id)]

  plot_save_expression_stats('vibrato_extend', note_params_reduced,
                             show_plot=show_plot,
                             save_fig=save_fig, output_dir=output_dir)

  plot_save_expression_stats('amplitude_mean', note_params_reduced,
                             show_plot=show_plot,
                             save_fig=save_fig, output_dir=output_dir)

  plot_save_expression_stats('amplitude_std', note_params_reduced,
                             show_plot=show_plot,
                             save_fig=save_fig, output_dir=output_dir)

  plot_save_expression_stats('brightness', note_params_reduced,
                             show_plot=show_plot,
                             save_fig=save_fig, output_dir=output_dir)

  plot_save_expression_stats('attack_level', note_params_reduced,
                             show_plot=show_plot,
                             save_fig=save_fig, output_dir=output_dir)

  plot_save_expression_stats('note_pitch', note_params_reduced,
                             show_plot=show_plot,
                             save_fig=save_fig, output_dir=output_dir)

  plot_save_expression_stats('note_length', note_params_reduced,
                             show_plot=show_plot,
                             save_fig=save_fig, output_dir=output_dir)


def preprocess_tensors(data):
  data_out = {}
  data_out['note_pitch'] = tf.convert_to_tensor(
    np.array([n['note_pitch'] for n in data]))
  data_out['note_length'] = tf.convert_to_tensor(
    np.array([n['note_length'] for n in data]))
  data_out['conditioning_feature'] = tf.convert_to_tensor(
    np.array([n['conditioning_feature'] for n in data]))
  data_out['instrument_id'] = tf.convert_to_tensor(
    np.array(data[0]['instrument_id']))
  return data_out


def make_dataset_no_segment(note_features_all):
  dataset_not_segmented = []
  for note_feature_piece in note_features_all:
    input_dict_piece = [note_features_to_input_dict(n) for n in
                        note_feature_piece]
    dataset_not_segmented.append(preprocess_tensors(input_dict_piece))
  return dataset_not_segmented


def make_and_save_pickles(data_dir, split, pickle_output_dir,
                          stats_plot_output_dir):
  split_stats_plot_output_dir = os.path.join(stats_plot_output_dir, f'{split}')
  os.makedirs(split_stats_plot_output_dir, exist_ok=True)
  note_features_all, audio_all = get_all_note_features(data_dir)
  note_features_all_reduced = []
  for n in note_features_all:
    note_features_all_reduced.extend(n)
  plot_stats(note_features_all_reduced, show_plot=False, save_fig=True,
             output_dir=split_stats_plot_output_dir)
  data_language_model = segment_data(note_features_all)
  data_language_model_preprocessed = [preprocess_tensors(d) for d in
                                      data_language_model]
  data_language_model_separate_piece = [segment_data([n]) for n in
                                        note_features_all]
  data_language_model_separate_piece_preprocessed = []
  for data in data_language_model_separate_piece:
    data_language_model_separate_piece_preprocessed.append(
      [preprocess_tensors(d) for d in data])
  not_segmented = make_dataset_no_segment(note_features_all)
  print('Total number of training pieces:', len(note_features_all))
  print('Total number of training samples:',
        len(data_language_model_preprocessed))

  pickle_dump(data_language_model_preprocessed,
              os.path.join(pickle_output_dir, f'{split}.pickle'))
  pickle_dump(data_language_model_separate_piece_preprocessed,
              os.path.join(pickle_output_dir, f'{split}_separate_piece.pickle'))
  pickle_dump(not_segmented,
              os.path.join(pickle_output_dir, f'{split}_not_segmented.pickle'))
  pickle_dump(audio_all,
              os.path.join(pickle_output_dir, f'{split}_audio_all.pickle'))


def dump_expression_generator_dataset(model, data_dir, output_dir):
  # TODO: (yusongwu) add automatic note expression scaling

  synth_params_output_dir = os.path.join(output_dir, 'synth_params')
  pickle_output_dir = os.path.join(output_dir, 'pickles')
  stats_plot_output_dir = os.path.join(output_dir, 'stats_plot')
  os.makedirs(pickle_output_dir, exist_ok=True)
  os.makedirs(stats_plot_output_dir, exist_ok=True)

  test_data_loader = UrmpMidiUnsegmented(data_dir, instrument_key='all',
                                       split='test', suffix='unbatched')
  test_dataset = test_data_loader.get_batch(batch_size=1, shuffle=True,
                                            repeats=1)
  train_data_loader = UrmpMidiUnsegmented(data_dir, instrument_key='all',
                                        split='train', suffix='unbatched')
  train_dataset = train_data_loader.get_batch(batch_size=1, shuffle=True,
                                              repeats=1)

  train_set_output_dir = os.path.join(synth_params_output_dir, 'train')
  os.makedirs(train_set_output_dir, exist_ok=True)
  generate_and_save_synth_params(model, train_dataset, train_set_output_dir)
  test_set_output_dir = os.path.join(synth_params_output_dir, 'test')
  os.makedirs(test_set_output_dir, exist_ok=True)
  generate_and_save_synth_params(model, test_dataset, test_set_output_dir)

  make_and_save_pickles(train_set_output_dir, 'train', pickle_output_dir,
                        stats_plot_output_dir)
  make_and_save_pickles(test_set_output_dir, 'test', pickle_output_dir,
                        stats_plot_output_dir)



#  Copyright 2022 The MIDI-DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Download the model weights"""

import zipfile
import os

try:
  from urllib.request import urlretrieve
except ImportError:
  from urllib import urlretrieve


def main():
  package_dir = os.path.dirname(os.path.realpath(__file__))
  weight_file_name = 'midi_ddsp_model_weights_urmp_9_10.zip'
  weight_file_download_path = os.path.join(package_dir, f'{weight_file_name}')
  print('Downloading midi-ddsp model weight files')
  urlretrieve(
    f'https://github.com/magenta/midi-ddsp/raw/models/{weight_file_name}',
    weight_file_download_path)
  print('Decompressing ...')
  with zipfile.ZipFile(weight_file_download_path, 'r') as zip_ref:
    zip_ref.extractall(package_dir)
  print('Decompression complete')
  os.remove(weight_file_download_path)


if __name__ == '__main__':
  main()



#  Copyright 2022 The MIDI-DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The constants for handling instrument names, abbreviations, ids and midi
programs used in the URMP dataset."""

INST_NAME_TO_ABB_DICT = {
  'violin': 'vn',
  'viola': 'va',
  'cello': 'vc',
  'double bass': 'db',
  'flute': 'fl',
  'oboe': 'ob',
  'clarinet': 'cl',
  'saxophone': 'sax',
  'bassoon': 'bn',
  'trumpet': 'tpt',
  'horn': 'hn',
  'trombone': 'tbn',
  'tuba': 'tba',
  'guitar': 'gtr'
}

INST_ABB_TO_NAME_DICT = {v: k for k, v in INST_NAME_TO_ABB_DICT.items()}

INST_ABB_LIST = list(INST_NAME_TO_ABB_DICT.values())

INST_NAME_LIST = list(INST_NAME_TO_ABB_DICT.keys())

INST_ABB_TO_ID_DICT = {v: i for i, (k, v) in
                       enumerate(INST_NAME_TO_ABB_DICT.items())}

INST_ID_TO_ABB_DICT = {i: v for i, (k, v) in
                       enumerate(INST_NAME_TO_ABB_DICT.items())}

INST_NAME_TO_ID_DICT = {k: i for i, (k, v) in
                        enumerate(INST_NAME_TO_ABB_DICT.items())}

INST_ID_TO_NAME_DICT = {i: k for i, (k, v) in
                        enumerate(INST_NAME_TO_ABB_DICT.items())}

NUM_INST = 20

# MIDI program number (1-128): https://en.wikipedia.org/wiki/General_MIDI
INST_NAME_TO_MIDI_PROGRAM_DICT = {
  'violin': 41,
  'viola': 42,
  'cello': 43,
  'double bass': 44,
  'flute': 74,
  'oboe': 69,
  'clarinet': 72,
  'saxophone': 67,
  'bassoon': 71,
  'trumpet': 57,
  'horn': 61,
  'trombone': 58,
  'tuba': 59,
  'guitar': 27
}

# To make the MIDI program number 0 indexed (0-127):
INST_NAME_TO_MIDI_PROGRAM_DICT = {k: v - 1 for k, v in
                                  INST_NAME_TO_MIDI_PROGRAM_DICT.items()}

INST_ABB_TO_MIDI_PROGRAM_DICT = {INST_NAME_TO_ABB_DICT[k]: v for k, v in
                                 INST_NAME_TO_MIDI_PROGRAM_DICT.items()}

INST_ID_TO_MIDI_PROGRAM_DICT = {INST_ABB_TO_ID_DICT[k]: v for k, v in
                                INST_ABB_TO_MIDI_PROGRAM_DICT.items()}

MIDI_PROGRAM_TO_INST_NAME_DICT = {v: k for k, v in
                                  INST_NAME_TO_MIDI_PROGRAM_DICT.items()}

MIDI_PROGRAM_TO_INST_ABB_DICT = {v: k for k, v in
                                 INST_ABB_TO_MIDI_PROGRAM_DICT.items()}

MIDI_PROGRAM_TO_INST_ID_DICT = {v: k for k, v in
                                INST_ID_TO_MIDI_PROGRAM_DICT.items()}



#  Copyright 2022 The MIDI-DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Model class for DDSP Inference module used in MIDI-DDSP."""

import tensorflow as tf
import ddsp
from midi_ddsp.utils.audio_io import tf_log_mel
from midi_ddsp.data_handling.instrument_name_utils import NUM_INST
from ddsp.training import nn
from ddsp.spectral_ops import F0_RANGE, DB_RANGE

tfk = tf.keras
tfkl = tfk.layers


class MelF0LDEncoder(tfkl.Layer):
  """The encoder in DDSP Inference.
  The MelF0LDEncoder takes input of audio, loudness and f0.
  The MelF0LDEncoder extract features from audio using an 8-layer CNN,
  and extract features from loudness and f0 using fully-connected layers.
  Then, a bi-lstm is used to extract contextual features from the extracted
  features.
  """

  def __init__(self, cnn, nhid, sample_rate, win_length, hop_length, n_fft,
               num_mels, fmin):
    super().__init__()
    self.nhid = nhid
    self.cnn = cnn
    self.z_fc = tfkl.Dense(nhid)
    self.f0_ld_fc = tfkl.Dense(nhid)
    self.rnn = tfkl.Bidirectional(
      tfkl.LSTM(units=nhid, return_sequences=True), name='bilstm'
    )
    # TODO(yusongwu): change emb dim to 64
    self.instrument_emb = tfkl.Embedding(NUM_INST, 128)

    # mel-spec parameters
    self.sample_rate = sample_rate
    self.win_length = win_length
    self.hop_length = hop_length
    self.n_fft = n_fft
    self.num_mels = num_mels
    self.fmin = fmin

  def call(self, inputs, training=False):
    mel = tf_log_mel(inputs['audio'],
                     self.sample_rate,
                     self.win_length,
                     self.hop_length,
                     self.n_fft,
                     self.num_mels,
                     self.fmin)
    z_cnn = self.cnn(mel, training=training)
    z_reduce = self.z_fc(z_cnn)
    instrument_z = tf.tile(
      self.instrument_emb(inputs['instrument_id'])[:, tf.newaxis, :],
      [1, z_cnn.shape[1], 1])
    x = tf.concat([ddsp.core.hz_to_midi(inputs['f0_hz']) / F0_RANGE,
                   inputs['loudness_db'] / DB_RANGE], -1)
    x_z = self.f0_ld_fc(x)
    z_out = self.rnn(tf.concat([x_z, z_reduce, instrument_z], -1))
    return z_out


class FCHarmonicDecoder(tfkl.Layer):
  """The decoder in DDSP Inference.
  The FCHarmonicDecoder takes input of a feature sequence,
  and output the synthesis parameters for DDSP through fully-connected layers.
  """

  def __init__(self, nhramonic=100, nnoise=65):
    super().__init__()

    self.harmonic_amp_fc = tfkl.Dense(1, bias_initializer='ones')
    self.harmonic_distribution_fc = tfkl.Dense(nhramonic)

    self.noise_mag_fc = tfkl.Dense(nnoise)

  def get_synth_params(self, inputs):
    z, data = inputs

    harmonic_amp = self.harmonic_amp_fc(z)
    harmonic_distribution = self.harmonic_distribution_fc(z)
    noise_mag = self.noise_mag_fc(z)

    synth_params = {
      'f0_hz': data['f0_hz'],
      'amplitudes': harmonic_amp,
      'harmonic_distribution': harmonic_distribution,
      'noise_magnitudes': noise_mag,
    }

    return synth_params

  def call(self, inputs):
    synth_params = self.get_synth_params(inputs)

    return synth_params


class F0LDEncoder(tfkl.Layer):
  """The encoder of original DDSP autoencoder."""

  # TODO: (yusongwu) To be removed and use the decoders.RnnFcDecoder
  def __init__(self):
    super().__init__()
    self.nhid = 512
    self.f0_fc = nn.FcStack(self.nhid, layers=3)
    self.ld_fc = nn.FcStack(self.nhid, layers=3)
    self.instrument_emb = tfkl.Embedding(NUM_INST, 128)
    self.rnn = tfkl.GRU(
      units=self.nhid, return_sequences=True,  # dropout=0.2,
    )

  def call(self, inputs, training=False):
    z_f0 = self.f0_fc(ddsp.core.hz_to_midi(inputs['f0_hz']) / F0_RANGE)
    z_ld = self.ld_fc(inputs['loudness_db'] / DB_RANGE)
    instrument_z = tf.tile(
      self.instrument_emb(inputs['instrument_id'])[:, tf.newaxis, :],
      [1, z_ld.shape[1], 1])
    x_z = tf.concat([z_f0, z_ld, instrument_z], -1)
    z_out = self.rnn(x_z)
    z_out = tf.concat([x_z, z_out], -1)
    return z_out


class FCStackHarmonicDecoder(tfkl.Layer):
  """The decoder original DDSP autoencoder.
  The FCStackHarmonicDecoder takes input of a feature sequence,
  and output the synthesis parameters for DDSP through stacked MLP.
  """

  def __init__(self, nharmonic=100, nnoise=65):
    super().__init__()

    self.output_splits = (
      ('amplitudes', 1), ('harmonic_distribution', nharmonic),
      ('noise_magnitudes', nnoise))
    self.n_out = sum([v[1] for v in self.output_splits])
    self.out_stack = nn.FcStack(512, layers=3)
    self.dense_out = tfkl.Dense(self.n_out)

  def get_synth_params(self, inputs):
    z, data = inputs

    z_output = self.out_stack(z)
    synth_params = nn.split_to_dict(self.dense_out(z_output),
                                    self.output_splits)

    synth_params['f0_hz'] = data['f0_hz']

    return synth_params

  def call(self, inputs):
    synth_params = self.get_synth_params(inputs)
    return synth_params


class ConvBlock(tfkl.Layer):
  """
  A tensorflow implementation of ConvBlock used in audioset classification.
  This CNN has better performance when used in spectrogram feature
  extraction for audio tagging. Adapted from pytorch implementation:
  https://github.com/qiuqiangkong/audioset_tagging_cnn.
  paper: https://arxiv.org/abs/1912.10211.
  Args:
    out_channels: number of output channels.
    pool_size: size of pooling, in height and width.
  """

  def __init__(self, out_channels, pool_size=(2, 2)):
    super().__init__()

    self.conv1 = tfkl.Conv2D(filters=out_channels,
                             kernel_size=(3, 3), strides=(1, 1),
                             padding='same', use_bias=False,
                             kernel_initializer=
                             tf.keras.initializers.GlorotUniform())

    self.conv2 = tfkl.Conv2D(filters=out_channels,
                             kernel_size=(3, 3), strides=(1, 1),
                             padding='same', use_bias=False,
                             kernel_initializer=
                             tf.keras.initializers.GlorotUniform())

    self.bn1 = tfkl.BatchNormalization(beta_initializer='zeros',
                                       gamma_initializer='ones')
    self.bn2 = tfkl.BatchNormalization(beta_initializer='zeros',
                                       gamma_initializer='ones')

    self.max_pool = tfkl.MaxPool2D(pool_size=pool_size, padding='same')
    self.avg_pool = tfkl.AveragePooling2D(pool_size=pool_size, padding='same')

  def call(self, inputs, training=None, pool_type='avg'):
    x = inputs
    x = tf.nn.relu(self.bn1(self.conv1(x), training=training))
    x = tf.nn.relu(self.bn2(self.conv2(x), training=training))
    if pool_type == 'max':
      x = self.max_pool(x)
    elif pool_type == 'avg':
      x = self.avg_pool(x)
    elif pool_type == 'avg+max':
      x1 = self.avg_pool(x)
      x2 = self.max_pool(x)
      x = x1 + x2
    else:
      raise Exception('Incorrect argument!')

    return x


class Cnn8(tfkl.Layer):
  """
  A tensorflow implementation of CNN8 used in audioset classification.
  This CNN has better performance when used in spectrogram feature
  extraction for audio tagging. Adapted from pytorch implementation:
  https://github.com/qiuqiangkong/audioset_tagging_cnn.
  paper: https://arxiv.org/abs/1912.10211.
  """

  def __init__(self, pool_size=(2, 2), dropout=0.2):
    super().__init__()

    self.conv_block1 = ConvBlock(out_channels=64, pool_size=pool_size)
    self.conv_block2 = ConvBlock(out_channels=128, pool_size=pool_size)
    self.conv_block3 = ConvBlock(out_channels=256, pool_size=pool_size)
    self.conv_block4 = ConvBlock(out_channels=512, pool_size=pool_size)

    self.dropout = tfkl.Dropout(rate=dropout)

  def call(self, x, training=None):
    x = x[..., tf.newaxis]

    x = self.conv_block1(x, pool_type='avg', training=training)
    x = self.dropout(x, training=training)
    x = self.conv_block2(x, pool_type='avg', training=training)
    x = self.dropout(x, training=training)
    x = self.conv_block3(x, pool_type='avg', training=training)
    x = self.dropout(x, training=training)
    x = self.conv_block4(x, pool_type='avg', training=training)
    x = self.dropout(x, training=training)
    x = tf.reshape(x, [x.shape[0], x.shape[1], -1])

    return x
